/* -*- Mode: C; tab-width: 4; indent-tabs-mode: nil; c-basic-offset: 2 -*-
 *
 * The contents of this file are subject to the Mozilla Public 
 * License version 1.1 (the "License"); you may not use this file
 * except in compliance with the License. You may obtain a copy of
 * the License at http://www.mozilla.org/MPL/
 *
 * Software distributed under the License is distributed on an "AS
 * IS" basis, WITHOUT WARRANTY OF ANY KIND, either express or 
 * implied. See the License for the specific language governing
 * rights and limitations under the License.
 *
 * The Initial Developer of the Original Code is R.V.Guha. Portions
 * created by R.V.Guha are copyright (c) 2000 R.V.Guha. All Rights
 * Reserved.
 *
 * Contributors : Jeen Broekstra, Jasper van de Gronde (win32 port)
 */


#include "dbglue.h"
#include <db.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#ifdef WIN32
#include <windows.h>
#include <io.h>
#else
#include <dirent.h>
#endif
#include "../utils/hash.h"
#include "rdf.h"
#include "../utils/utils.h"
#include <pthread.h>

typedef struct _IndexStruct* Index;
#include "indexPool.h"
#define MAX_ANS_SIZE 4096

#define ALL_PERMS (S_IRWXU | S_IRWXO | S_IRWXG)

/* JvdG Made this enum in the process of adding support for EVERYTHING_QUERY. */
typedef enum {
  KEY_QUERY,        /**< Just the key is important. */
  KEY_VALUE_QUERY,  /**< Both the key and value are important (used for GetArcs*). */
  EVERYTHING_QUERY, /**< Just return every triple there is. */
} DB_QueryType;

typedef struct _DB_CursorStruct {
  DBT key;
  DBT data;
  DBC *cursor;
  DB  *db;
  int inited;
  char* source; /* JvdG Used only for EVERYTHING_QUERY. */
  char* arc; /* JvdG Used only for EVERYTHING_QUERY. */
  void* value;
  size_t valueSize;
  int type;
  int count;
  DB_QueryType query_type; /* JvdG Used to be int qtype; */
  const char* ind;
  struct _DB_CursorStruct *next; /**< Used for the cursor pool. */
  char ans[MAX_ANS_SIZE+1];
  char ans2[MAX_ANS_SIZE+1]; /* JvdG */
} DB_CursorStruct;

typedef struct _BDB_FileStruct {
  char* url;
  char* name;
  struct _BDB_FileStruct* next;
} BDB_FileStruct;

typedef BDB_FileStruct* BDB_File;

typedef struct _BDB_Struct {
  char* name; 
  int    nextAssertionRecord;
  DB*    findex;
  DB*    rindex;
  DB*    arcindex;
  BDB_File files;
} BDB_struct;

typedef BDB_struct* BDB;

static HashTable gDBHash = 0;

static char* gDatabaseDirectory = 0;

static pthread_mutex_t gOpenDBMutex = PTHREAD_MUTEX_INITIALIZER;

static DB_Cursor gDBCursorPool = 0;
static pthread_mutex_t gDBCursorPoolMutex = PTHREAD_MUTEX_INITIALIZER;

/*DB_ENV* gdb_env = 0;*/

int 
DB_Initialize() {
  if (!gDBHash) gDBHash = NewHashTable(256);
  if (!gDatabaseDirectory) {
    gDatabaseDirectory = strdup(get_setting("DIR"));
    if (!gDatabaseDirectory || strlen(gDatabaseDirectory) == 0) { /* JvdG Added the strlen check */
      PrintMsg(CRITICAL, "\n********* Please set the (environment) variable (RDFDB_)DIR to the directory where you want the data to be stored!**********\ne.g., export RDFDB_DIR=/var/rdfdb/\n");
      return 1;
    }
    else if (0 != chdir(gDatabaseDirectory)) { /* JvdG Added this check */
      PrintMsg(CRITICAL, "\nRDFDB_DIR isn't set to a valid directory!\n********* Please set the (environment) variable (RDFDB_)DIR to the directory where you want the data to be stored!**********\ne.g., export RDFDB_DIR=/var/rdfdb/\n");
      return 1;
    }
    else {
      /* JvdG Made sure the path always ends with a slash or backslash */
#ifdef WIN32
      if ( gDatabaseDirectory[strlen(gDatabaseDirectory)-1] != '\\' ) {
        gDatabaseDirectory = (char*)realloc(gDatabaseDirectory, strlen(gDatabaseDirectory)+2);
        strcat(gDatabaseDirectory, "\\");
      }
#else
      if ( gDatabaseDirectory[strlen(gDatabaseDirectory)-1] != '/' ) {
        gDatabaseDirectory = (char*)realloc(gDatabaseDirectory, strlen(gDatabaseDirectory)+2);
        strcat(gDatabaseDirectory, "/");
      }
#endif
      PrintMsg(INFO, "Database directory = %s\n", gDatabaseDirectory);
    }
  }
  return 0;
}

static void CloseAllDatabases();
static void EmptyCursorPool();

/* JvdG */
int 
DB_Deinitialize() {
  EmptyCursorPool();
  CloseAllDatabases();
  if ( gDBHash ) {
    DeleteHashTable(gDBHash);
    gDBHash = 0;
  }
  if ( gDatabaseDirectory ) {
    free(gDatabaseDirectory);
    gDatabaseDirectory = 0;
  }
  return 0;
}


DB_Cursor
getNewDBCursor () {
  DB_Cursor existing;
  pthread_mutex_lock(&gDBCursorPoolMutex);
  existing = gDBCursorPool;
  if (existing) {
    gDBCursorPool = existing->next;
    bzero(existing, sizeof(DB_CursorStruct));
  } 
  pthread_mutex_unlock(&gDBCursorPoolMutex);
  if (!existing) {
    existing = (DB_Cursor) getMem(sizeof(DB_CursorStruct));
  }
  return existing;
}

void 
freeDBCursor (DB_Cursor c) {
  pthread_mutex_lock(&gDBCursorPoolMutex);
  c->next = gDBCursorPool;
  gDBCursorPool = c;
  pthread_mutex_unlock(&gDBCursorPoolMutex);
}

/* Jasper van de Gronde
 * Changed the first argument of the third ->close to bdb->arcindex (it used to be bdb->rindex)
 * Added freeMem(bdb).
 */
static
void
CloseAllDatabases() {
  int i;
  for(i = 0; i < gDBHash->size; ++i) {
    HashEntry b = gDBHash->buckets[i];
    while (b) {
      BDB bdb = b->value;
      bdb->findex->close(bdb->findex, 0);
      bdb->rindex->close(bdb->rindex, 0);
      bdb->arcindex->close(bdb->arcindex, 0);
      freeMem(bdb);
      b = b->next;
    }
  }
}

/* JvdG */
static void EmptyCursorPool() {
  DB_Cursor next;

  pthread_mutex_lock(&gDBCursorPoolMutex);

  while(gDBCursorPool) {
    next = gDBCursorPool->next;
    freeMem(gDBCursorPool);
    gDBCursorPool = next;
  }

  pthread_mutex_unlock(&gDBCursorPoolMutex);
}

/*
int DB_DatabaseExistsp (const char* db) {
  char fullName[256];
#ifdef WIN32
  struct _stat stats;
#else
  DIR *dir = 0;
#endif
  snprintf(fullName, 256, "%s%s", gDatabaseDirectory, db);
#ifdef WIN32
  if ( _stat(fullName, &stats) != 0 ) return 0;
#else
  dir = opendir(fullName);
  if (!dir) return 0;
  closedir(dir);
#endif
  return 1;
}*/

/*
void
init_db_env () {
  db_env_create(&gdb_env, 0);
  gdb_env->open(gdb_env, gDatabaseDirectory, 0, 0,0);
  } */

int
openTable (const char* dbName, DB** db, const char* name, int type,  int dups) {
  int err = db_create(db, 0, 0);

  if (err) {
    PrintMsg(CRITICAL, "Problems creating %s : %s\n", name, db_strerror(err));
    return err;
  }
  
  if (dups) {
    err = (*db)->set_flags(*db, DB_DUP);
    if (err) {
      PrintMsg(CRITICAL, "Problems setting duplicate flag for %s : %s\n", name, db_strerror(err));
      return err;
    }
  }

  err = (*db)->open(*db, NULL /* Added, txn_id */, name, NULL, type, DB_CREATE | DB_THREAD , 0664);
  if (err) {
    PrintMsg(CRITICAL, "Problems opening %s : %s\n", name, db_strerror(err));
    return err;
  }
  return 0;
}

#ifdef WIN32
void
initBDBFiles (BDB db) {
  char fullName[256];
  char *buff = getMem(1024);
  HANDLE find_handle=NULL;
  WIN32_FIND_DATA find_data;
  snprintf(fullName, 256, "%s%s\\*.dbfl", gDatabaseDirectory, db->name);
  find_handle = FindFirstFile(fullName, &find_data);
  while(find_handle!=INVALID_HANDLE_VALUE)
  {
    char* name = find_data.cFileName;
    char* suffix = strchr(name, '.');
    BDB_File dbf = (BDB_File) getMem(sizeof(BDB_FileStruct));
    int fi = open(name, O_RDONLY);
    char *end;
    dbf->name = getMem(suffix-name+1);
    memcpy(dbf->name, name, suffix-name); /* JvdG Used to be suffix-name-1 */
    dbf->next = db->files;
    db->files = dbf;
    bzero(buff, 1024);
    read(fi, buff, 1024);
    end = strchr(buff, '\n');
    if (end) *end = 0;
    dbf->url = strdup(buff);
    close(fi);
    if ( !FindNextFile(find_handle, &find_data) ) break;
  }
  FindClose(find_handle);
  freeMem(buff);
}
#else
void
initBDBFiles (BDB db) {
  char fullName[256];
  char *buff = getMem(1024);
  DIR *d;
  struct dirent *f;
  snprintf(fullName, 256, "%s%s", gDatabaseDirectory, db->name);
  d = opendir(fullName);
  if (!d) {
    freeMem(buff);
    return;
  }
  while (0 != (f = readdir(d))) {
    char* name = f->d_name;
    char* suffix = strchr(name, '.');
    if (suffix && streq(suffix+1, "dbfl")) {
      BDB_File dbf = (BDB_File) getMem(sizeof(BDB_FileStruct));
      int fi = open(name, O_RDONLY);
      char *end;
      dbf->name = getMem(suffix-name+1);
      memcpy(dbf->name, name, suffix-name); /* JvdG Used to be suffix-name-1 */
      dbf->next = db->files;
      db->files = dbf;
      bzero(buff, 1024);
      read(fi, buff, 1024);
      end = strchr(buff, '\n');
      if (end) *end = 0;
      dbf->url = strdup(buff);
      close(fi);
    }
  }
  closedir(d);
  freeMem(buff);
}
#endif

int
openDB (const char* name, int createp) {
  char fullName[256];
  BDB db  = (BDB)getMem(sizeof(BDB_struct));
  int err =0;
  snprintf(fullName, 256, "%s%s", gDatabaseDirectory, name);
  if (chdir(fullName)) {
    if (createp) {
#ifdef WIN32
      if (0 != (err = mkdir(fullName))) {
#else
      if (0 != (err = mkdir(fullName, ALL_PERMS))) {
#endif
        PrintMsg(CRITICAL, "Cannot mkdir  %s\n", fullName);
        return RDF_WRONG_FILE_PERMISSIONS_ERROR;
      }
      if (0 != (err = chdir(fullName))) {
        PrintMsg(CRITICAL, "Cannot chdir to %s\n", fullName);
        return RDF_WRONG_FILE_PERMISSIONS_ERROR;
      }
    } else {
      PrintMsg(CRITICAL, "%s does not exist!\n", fullName);
      return RDF_DB_DOES_NOT_EXIST;
    }
  }
  db->name = strdup(name);
  if (0 != (err = openTable(name, &(db->findex), "findex.db", DB_BTREE, 1))) {
    PrintMsg(CRITICAL, "Cannot open findex.db\n");
    return err;
  }
  if (0 != (err = openTable(name, &(db->rindex), "rindex.db", DB_BTREE, 1))) {
    PrintMsg(CRITICAL, "Cannot open rindex\n");
    return err;
  }
  if (0 != (err = openTable(name, &(db->arcindex), "arcindex.db", DB_BTREE, 1))) {
    PrintMsg(CRITICAL, "Cannot open arcindex \n");
    return err;
  }

  initBDBFiles(db);
  HashAdd(gDBHash, name, db);
  return 0;
}


/* JvdG
 * Changed this to not always create the db (this used to be called DB_CreateDatabase and had a parameter less).
 * Made it lock gOpenDBMutex.
 */
int 
DB_OpenDatabase (const char* db, int create) {
  int ret = 0;
  pthread_mutex_lock(&gOpenDBMutex);
  if (!HashLookup(gDBHash, db)) {
    ret = openDB(db, create);
  }
  pthread_mutex_unlock(&gOpenDBMutex);
  return ret;
}

/* JvdG Restructured this to be clearer and more correct. */
BDB
db2bdb (const char* name, int createp) {
  BDB bdb;
  pthread_mutex_lock(&gOpenDBMutex);
  bdb = HashLookup(gDBHash, name);
  if ( !bdb ) {
    openDB(name, createp);
    bdb = HashLookup(gDBHash, name);
  }
  pthread_mutex_unlock(&gOpenDBMutex);
  return bdb;
}

int
DB_GetFileCode (const char* db, const char *url, char **code, FILE **file, int open_for_write) {
  BDB bdb = db2bdb(db, 0);
  BDB_File bf ;
  int n = 1;
  char fullName[256];
  char lbuff[512];

  if (!bdb) return 0;
  bf = bdb->files;
  pthread_mutex_lock(&gOpenDBMutex);
  while (bf) {
    if (strcasecmp(bf->url, url) == 0) {
      pthread_mutex_unlock(&gOpenDBMutex);
      snprintf(fullName, 256, "%s%s/%s.dbfl", gDatabaseDirectory, bdb->name, bf->name); /* JvdG Made this use bf->name instead of n. */
      if ( open_for_write ) {
        *file = fopen(fullName, "a+");
      }
      else {
        *file = fopen(fullName, "r+");
        if (*file) {
          fgets(lbuff, 512, *file);
        }
      }
      *code = bf->name;
      return 0;
    }
    bf = bf->next;
    n++;
  }

  if ( !open_for_write ) return 0;

  bf = (BDB_File)getMem(sizeof(BDB_FileStruct));
  bf->next = bdb->files;
  bdb->files = bf;
  
  bf->url = strdup(url);
  bf->name = getMem(12);
  snprintf(bf->name, 12, "%i", n);

  snprintf(fullName, 256, "%s%s/%i.dbfl", gDatabaseDirectory, bdb->name, n);
  *file = fopen(fullName, "a+");
#ifdef WIN32
#else
  chmod(fullName, ALL_PERMS);
#endif

  if (*file) fprintf(*file, "%s\n", url);

  *code = bf->name;
   
  pthread_mutex_unlock(&gOpenDBMutex);
  return 0;
}
  
/* not used */
/*
int 
DB_CloseDatabase (const char* dbName) {
  BDB bdb = db2bdb(dbName, 0);
  bdb->findex->close(bdb->findex, 0);
  bdb->rindex->close(bdb->rindex, 0);
  bdb->arcindex->close(bdb->rindex, 0);
  return 0;
}
*/
/*
size_t itemLen (void* item, RDF_ValueType type) {
  if ((type == RDF_RESOURCE_TYPE) || (type == RDF_STRING_TYPE))  {
    return strlen((char*) item);
  } else return 12;
}
*/
/* JvdG Made this use DB_QueryType and added support for EVERYTHING_QUERY. */
int
getIndex (DB* db, const char* index, DB_Cursor *c, DB_QueryType query_type) {
  DBC *dbc;
  DB_Cursor xdbc = getNewDBCursor();
  int err = 0;

  err = db->cursor(db, 0, &dbc, 0);
  if (err) {
    freeMem((char*)xdbc);
    return err;
  }

  xdbc->query_type = query_type;
  switch(query_type) { /* JvdG Made this use a switch instead of an if. */
  case KEY_QUERY:
    xdbc->key.data = index;
    xdbc->key.size  = strlen(index);
    xdbc->data.data = xdbc->ans;
    xdbc->data.flags = DB_DBT_USERMEM;
    xdbc->data.ulen = MAX_ANS_SIZE;
    break;
  case KEY_VALUE_QUERY:
    xdbc->key.data = xdbc->ans;
    xdbc->key.size  = strlen(index);
    xdbc->ind = index;
    memcpy(xdbc->key.data, index, strlen(index));
    xdbc->key.flags = DB_DBT_USERMEM;
    xdbc->key.ulen = MAX_ANS_SIZE;
    break;
  case EVERYTHING_QUERY:
    xdbc->key.data = xdbc->ans;
    xdbc->key.flags = DB_DBT_USERMEM;
    xdbc->key.ulen = MAX_ANS_SIZE;
    xdbc->data.data = xdbc->ans2;
    xdbc->data.flags = DB_DBT_USERMEM;
    xdbc->data.ulen = MAX_ANS_SIZE;
    break;
  }
  xdbc->cursor = dbc;
  xdbc->db = db;
  *c = xdbc;
  return 0;

}

/* JvdG Changed this to precalculate strlen(value) and use it in the duplicate check, otherwise it mistakenly thought ...MIT was the same as ...MIT2.
*/
int
addremoveIndex (DB* db, const char* index, const char* value, int op, int dupCheck) {
  static int asc = 0;
  static int nasc = 0;

  DBC *dbc;
  DBT key;
  DBT data;
  const size_t value_len = strlen(value);
  int err = 0;
  int found = 0;
  int first = 1;
  int somethingFound = 0;
  bzero(&key, sizeof(DBT));
  bzero(&data, sizeof(DBT));

  db->cursor(db, 0, &dbc, 0);

  key.data = index;
  key.size  = strlen(index);

  data.data = value;
  data.size  = value_len;

  /*  if (dupCheck)   printf("\n***********************\n"); */

  if (op || dupCheck) {
    for (;;) {
      int err = dbc->c_get(dbc, &key, &data, (first ? DB_SET : DB_NEXT_DUP));
      first = 0;
      if (err) {
        if (err == DB_NOTFOUND) {
          break;
        } else {
          PrintMsg(CRITICAL, "Problems getting cursor during %s for key=%s, value=%s: %s\n", 
                 (op ? "removal" : "addition"), index, value, db_strerror(err));
          return err;
        }
      }
      somethingFound = 1;
      /*    printf("%s ===> %s\n", (char*)key.data, &((char*)data.data)[4]); */
      if (data.size == value_len && strncmp((char*)(data.data), value, data.size) == 0) {
        if (op == 1) {
          err = dbc->c_del(dbc, 0);
          if (err) {
            PrintMsg(CRITICAL, "Problems removing key=%s, value=%s: %s\n", 
                   index, value, db_strerror(err));
            dbc->c_close(dbc);
            return err;
          }
          found = 1;
        } else {
          found = 1;
          PrintMsg(FAILURE, "\tAlready in database\n");
        }
        break;
      }
    }
  }
  if ((op == 0) && (!found)) {

    bzero(&data, sizeof(DBT));
    bzero(&data, sizeof(DBT));
    key.data = index;
    key.size  = strlen(index);
    data.data = value;
    data.size  = value_len;
    err = dbc->c_put(dbc, &key, &data, DB_KEYFIRST);    
    asc++;
    if (asc == 1000) {
      nasc++;
      PrintMsg(INFO, "[%i]", nasc);
      asc = 0;
    }
    if (err) {
      PrintMsg(CRITICAL, "Problems putting  for key=%s, value=%s: %s\n", 
              index, value, db_strerror(err));
      dbc->c_close(dbc);
      return err;
    } else {
      PrintMsg(DEBUG_INFO, "\nAdded %s (key %s)\n", (char*)data.data, (char*)key.data);
    }
  }
  
  dbc->c_close(dbc);

  if (op == 0 && dupCheck) {
    if (found) {
      return 2;
    } else if (somethingFound) {
      return 1;
    }
  }

  /* JvdG Added this to make sure RDF_Unassert doesn't count non-existing records as being deleted. */
  if (op == 1 && !found) {
    return 2;
  }

  return 0;
}

int
DB_Flush (const char* db) {
  BDB bdb = db2bdb(db, 0);
  if (bdb) {
    bdb->findex->sync(bdb->findex, 0);
    bdb->rindex->sync(bdb->rindex, 0);
    bdb->arcindex->sync(bdb->arcindex, 0);
    return 0;
  } else return RDF_DB_DOES_NOT_EXIST;
}
  
int 
DB_AddForwardIndex(const char* dbName, const char* forwardIndex, const char* value, 
                   int dupCheck) {
  BDB db = db2bdb(dbName, 0);
  if (!db) return RDF_DB_DOES_NOT_EXIST;
  return addremoveIndex(db->findex, forwardIndex, value, 0, dupCheck);
}

int DB_RemoveForwardIndex(const char* dbName, const char* forwardIndex, const char* value) {
  BDB db = db2bdb(dbName, 0);
  if (!db) return RDF_DB_DOES_NOT_EXIST;
  return addremoveIndex(db->findex, forwardIndex, value, 1, 1);
}

int DB_AddReverseIndex(const char* dbName, const char* reverseIndex, const char* value, int dupCheck) {
  BDB db = db2bdb(dbName, 0);
  if (!db) return RDF_DB_DOES_NOT_EXIST;
  return addremoveIndex(db->rindex, reverseIndex, value, 0, dupCheck);
}

int DB_RemoveReverseIndex(const char* dbName, const char* reverseIndex, const char* value) {
  BDB db = db2bdb(dbName, 0);
  if (!db) return RDF_DB_DOES_NOT_EXIST;
  return addremoveIndex(db->rindex, reverseIndex, value, 1, 1);
}


int DB_AddArcIndex(const char* dbName, const char* arcIndex, const char* value, 
                   int dupCheck) {
  BDB db = db2bdb(dbName, 0);
  if (!db) return RDF_DB_DOES_NOT_EXIST;
  return addremoveIndex(db->arcindex, arcIndex, value, 0, dupCheck);
}

int DB_RemoveArcIndex(const char* dbName, const char* arcIndex, const char* value) {
  BDB db = db2bdb(dbName, 0);
  if (!db) return RDF_DB_DOES_NOT_EXIST;
  return addremoveIndex(db->arcindex, arcIndex, value, 1, 1);
}


int
DB_GetForwardCursor(const char* dbName, const char* forwardIndex, DB_Cursor *c) {
  BDB db = db2bdb(dbName, 0);
  if (!db) return RDF_DB_DOES_NOT_EXIST;
  return getIndex(db->findex, forwardIndex, c, KEY_QUERY);
}

int
DB_GetReverseCursor(const char* dbName, const char* reverseIndex, DB_Cursor *c) {
  BDB db = db2bdb(dbName, 0);
  if (!db) return RDF_DB_DOES_NOT_EXIST;
  return getIndex(db->rindex, reverseIndex, c, KEY_QUERY);
}

int 
DB_GetArcsOutCursor(const char* dbName, const char* index, DB_Cursor *c) {
 BDB db = db2bdb(dbName, 0);
 if (!db) return RDF_DB_DOES_NOT_EXIST;
 return getIndex(db->findex, index, c, KEY_VALUE_QUERY);
}

int 
DB_GetArcsInCursor(const char* dbName, const char* index, DB_Cursor *c) {
 BDB db = db2bdb(dbName, 0);
 if (!db) return RDF_DB_DOES_NOT_EXIST;
 return getIndex(db->rindex, index, c, KEY_VALUE_QUERY);
}
 
int
DB_GetArcCursor(const char* dbName, const char* arcIndex, DB_Cursor *c) {
  BDB db = db2bdb(dbName, 0);
  if (!db) return RDF_DB_DOES_NOT_EXIST;
  return getIndex(db->arcindex, arcIndex, c, KEY_QUERY);
}

int 
DB_GetEverythingCursor(const char* dbName, DB_Cursor *c) {
 BDB db = db2bdb(dbName, 0);
 if (!db) return RDF_DB_DOES_NOT_EXIST;
 return getIndex(db->findex, 0, c, EVERYTHING_QUERY);
}

/* JvdG Made this use DB_QueryType and added support for EVERYTHING_QUERY. */
int
DB_CursorAdvance(DB_Cursor c) {
  int err;
  int flags;

  /* JvdG Replaced ifs by the if and switches below. */
  /*if (!c->qtype) {
    if (!c->inited) 
      flags = DB_SET;
    else
      flags = DB_NEXT_DUP;
  } else {
    if (!c->inited) 
      flags = DB_SET_RANGE;
    else
      flags = DB_NEXT;
  }    
  */
  if ( !c->inited ) {
    /* Have to find the key */
    switch(c->query_type) {
    case KEY_QUERY:         flags = DB_SET; break;
    case KEY_VALUE_QUERY:   flags = DB_SET_RANGE; break;
    case EVERYTHING_QUERY:  flags = DB_FIRST; break;
    }
  }
  else {
    /* Have to find the next record */
    switch(c->query_type) {
    case KEY_QUERY:         flags = DB_NEXT_DUP; break;
    case KEY_VALUE_QUERY:   flags = DB_NEXT; break;
    case EVERYTHING_QUERY:  flags = DB_NEXT; break;
    }
  }

  c->data.size = 0;
  err = c->cursor->c_get(c->cursor, &(c->key), &(c->data), flags);
  c->inited = 1;
  if (err == DB_NOTFOUND) {
    return RDF_NO_MORE_VALUES;
  } else if (err != 0) {
    PrintMsg(CRITICAL, "Problem with query %s\n", db_strerror(err));
    return err;
  } else {
    switch(c->query_type) {
    case EVERYTHING_QUERY:
      c->source = c->key.data;
      c->arc = strchr(c->key.data, '|')+1; /* This works correctly because EVERYTHING_QUERY uses the forward index and its key is in the format Src|Arc and Src shouldn't contain a pipe (|). */
      *(c->arc-1) = 0;
      *(c->source+c->key.size) = 0;
      /* The lack of break here is intentional, EVERYTHING_QUERY just does a bit more work than KEY_QUERY. */
    case KEY_QUERY:
      {
        char* value = c->data.data;
        size_t len  = c->data.size;
        RDF_ValueType type = (RDF_ValueType) atoi(value);
        c->type = type;      
        c->count++;
        if ((type == RDF_RESOURCE_TYPE) || (type == RDF_STRING_TYPE)) {
          c->value = strchr(&value[2], '\xA')+1;
          c->valueSize = len - ((char*)c->value - value);
          return 0;
        } else if (type == RDF_INT_TYPE) {
          c->value = (void*)atoi(strchr(&value[2], '\xA')+1);
          c->valueSize = sizeof(int);
          return 0;
        }
      }
      break;
    case KEY_VALUE_QUERY:
      {
        char* value = c->key.data;
        size_t len  = c->key.size - strlen(c->ind)+1;
        if (!startsWith(value, c->ind)) return RDF_NO_MORE_VALUES;   
        c->type = RDF_RESOURCE_TYPE;     
        value[c->key.size] = 0; 
        if ((!startsWith(value, c->ind)) || (c->value && (streq(c->value, &value[strlen(c->ind)])))) {
          flags = DB_NEXT;
          for (;;) {
            c->data.size = 0;
            err = c->cursor->c_get(c->cursor, &(c->key), &(c->data), flags);
            if (err == DB_NOTFOUND) return RDF_NO_MORE_VALUES;
            if (err) return err;
            value = c->key.data;
            if (!startsWith(value, c->ind)) return RDF_NO_MORE_VALUES;   
            value[c->key.size] = 0;
            len  = c->key.size - strlen(c->ind)+1;
            if (!streq(c->value, &value[strlen(c->ind)])) break;
          }
        }
        c->valueSize = len;
        c->count++;
        if (c->value) freeMem(c->value); 
        c->value = getMem(strlen(value));
        memcpy(c->value, &value[strlen(c->ind)], len);
        /*    printf(">>>%s<<<%s>>>%s\n", (char*)c->value, value, c->ind); */
      }
      break;
    }
  }
  return 0;
}


int 
DB_CursorCurrentValue (DB_Cursor c, void** value) {
  if (!c) {
    return RDF_ERROR;
  } else {
    *value = c->value;
    return 0;
  }
}

int 
DB_CursorCurrentValueType (DB_Cursor c, int *type) {
  if (!c) {
    return RDF_ERROR;
  } else {
    *type = c->type;
    return 0;
  }
}

int 
DB_CursorCurrentValueSize (DB_Cursor c, int *sz) {
  if (!c) {
    return RDF_ERROR;
  } else {
    *sz = c->valueSize;
    return 0;
  }
}

/* JvdG */
int 
DB_CursorCurrentSource (DB_Cursor c, void** value) {
  if (!c || c->query_type != EVERYTHING_QUERY) {
    return RDF_ERROR;
  } else {
    *value = c->source;
    return 0;
  }
}

/* JvdG */
int 
DB_CursorCurrentArc (DB_Cursor c, void** value) {
  if (!c || c->query_type != EVERYTHING_QUERY) {
    return RDF_ERROR;
  } else {
    *value = c->arc;
    return 0;
  }
}

/* JvdG Made this use KEY_VALUE_QUERY instead of 2. */
int DB_ReleaseCursor(DB_Cursor c) {
  /*   char* keydata = c->key.data; */  
  c->cursor->c_close(c->cursor); 
  if (c->query_type == KEY_VALUE_QUERY ) {
    if (c->value) freeMem(c->value);
  }
  freeDBCursor(c); 
  return 0;
}








