/* -*- Mode: C; tab-width: 4; indent-tabs-mode: nil; c-basic-offset: 2 -*-
 *
 * The contents of this file are subject to the Mozilla Public 
 * License version 1.1 (the "License"); you may not use this file
 * except in compliance with the License. You may obtain a copy of
 * the License at http://www.mozilla.org/MPL/
 *
 * Software distributed under the License is distributed on an "AS
 * IS" basis, WITHOUT WARRANTY OF ANY KIND, either express or 
 * implied. See the License for the specific language governing
 * rights and limitations under the License.
 *
 * The Initial Developer of the Original Code is R.V.Guha. Portions
 * created by R.V.Guha are copyright (c) 2000 R.V.Guha. All Rights
 * Reserved.
 *
 * Contributors : Jasper van de Gronde (win32 port)
 */
/** \file */

#include "parsers.h"
#include "rdf.h"
#include "../utils/utils.h"
#include "rdf-int.h"
#include <stdio.h>
#include <raptor.h>

extern int gAssertionCount;

/* Data types */
#define XMLS_INTEGER "http://www.w3.org/2001/XMLSchema#integer"
raptor_uri* xmls_integer = 0;

/* JvdG */
int initParsers()
{
  int failures = 0;
  raptor_init();
  xmls_integer = raptor_new_uri(XMLS_INTEGER);
  if ( xmls_integer == 0 ) ++failures;
  return failures;
}

/* JvdG */
int deinitParsers()
{
  int failures = 0;
  raptor_free_uri(xmls_integer);
  raptor_finish();
  return failures;
}

void unloadExistingAssertions(FILE *f, const RDF_DB db, const char* sourceCode);

RDF_Result         
RDF_UnLoadURLFromDB (RDF_DB db, const char* url) {
  char *code;
  FILE *f = 0;
  DB_GetFileCode(db, url, &code, &f, 0);
  unloadExistingAssertions(f, db, code);
  if ( f ) fclose(f);
  return 0;
}

/* from librdf/db.c */
typedef void* BDB; /* not entirely correct, but it suffices for this purpose */
BDB db2bdb (const char* name, int createp);

static void RDF_StatementHandler(void *user_data, const raptor_statement *statement);
static unsigned char* RDF_GenerateIDHandler(void *user_data, raptor_genid_type type, unsigned char* user_bnodeid);

/* Jasper van de Gronde
 * Modified this to use raptor (see http://www.redland.opensource.ac.uk/raptor/).
 */
/** If code is null the triples will be added without a file code attached (and no undo file will be created).
 * \todo Handle parse errors.
 */
RDF_Result 
RDF_LoadURLIntoDB (RDF_DB db, const char* url, RDF_SerializationType type, 
                   RDF_Operation operation, char** code) {
  if (operation == RDF_UNASSERT && code) {
    return RDF_ERROR;
  }
  else if (db2bdb(db, 0) == 0) {
    return RDF_DB_DOES_NOT_EXIST;
  } else {
    RDF_Result ret_code = RDF_OK;
    RDF_Parser rdfp = (RDF_Parser)getMem(sizeof(RDF_ParserStruct));;

    rdfp->count = 0;
    rdfp->db = db;
    rdfp->url = url;
    if ( code ) {
      DB_GetFileCode(rdfp->db, rdfp->url, &rdfp->code, &rdfp->file, 1);
      *code = rdfp->code;
    }
    else {
      rdfp->code = 0;
      rdfp->operation = operation;
    }

    {
      raptor_parser *p;
      raptor_uri *uri=raptor_new_uri(url);
      
      if ( uri ) {
        char* nurl = (char*)malloc(2+strlen(url)+1+1);
        if ( nurl ) {
          strcpy(nurl, "_:");
          strcpy(nurl+2, url);
          strcat(nurl, "-");

          rdfp->blank_node_prefix = nurl;
          rdfp->blank_node_prefix_len = strlen(nurl);

          switch(type) {
          case RSS:
            p = raptor_new_parser("rss-tag-soup");
            break;
          case NTRIPLES:
            p = raptor_new_parser("ntriples");
            break;
          case TURTLE:
            p = raptor_new_parser("turtle");
            break;
          case RDF_XML:
          default:
            p = raptor_new_parser("rdfxml");
          }
          if ( p ) {
            raptor_set_statement_handler(p, rdfp, RDF_StatementHandler);
            /*JvdG Doesn't do anything for blank nodes with nodeIDs:
            raptor_set_default_generate_id_parameters(p, nurl, 1);*/
            raptor_set_generate_id_handler(p, rdfp, RDF_GenerateIDHandler);
            raptor_parse_uri(p, uri, 0);

            raptor_free_parser(p);
          }
          else { /* !p */
            ret_code = RDF_ERROR;
          }
          free(nurl);
        }
        else { /* !nurl */
          ret_code = RDF_ERROR;
        }
        raptor_free_uri(uri);
      }
      else { /* !uri */
        ret_code = RDF_ERROR;
      }
    }

    if ( rdfp->file ) fclose(rdfp->file);
    if ( rdfp ) freeMem((void*)rdfp);

    return ret_code;
  }
}

/* Jasper van de Gronde
 * Replacement of newAnonResource to be compatible with Raptor.
 */
static
unsigned char*
RDF_GenerateIDHandler(void *user_data, raptor_genid_type type, unsigned char* user_bnodeid) {
  RDF_Parser rdfp = (RDF_Parser)user_data;
  if ( user_bnodeid ) {
    const size_t len = rdfp->blank_node_prefix_len + strlen(user_bnodeid);
    unsigned char* const blank_node_id = malloc(len+1);
    if ( blank_node_id ) {
      strcpy(blank_node_id, rdfp->blank_node_prefix);
      strcpy(blank_node_id+rdfp->blank_node_prefix_len, user_bnodeid);
    }
    return blank_node_id;
  }
  else {
    const size_t numdigits = numDigits(++(rdfp->count));
    const size_t len = rdfp->blank_node_prefix_len + numdigits;
    unsigned char* const blank_node_id = malloc(len+1);
    if ( blank_node_id ) {
      strcpy(blank_node_id, rdfp->blank_node_prefix);
      snprintf(blank_node_id+rdfp->blank_node_prefix_len, numdigits, "%i", rdfp->count);
      blank_node_id[len] = 0; /* snprintf prints up to count characters, if the number of characters (not including the null terminator) stored is >= count it doesn't print a terminating null, so in any case it's safest to put the terminating null in myself. */
    }
    return blank_node_id;
  }
}
/* Jasper van de Gronde
 * I changed len below to be more exact, len used to be too small for counts above 99999.
 * I also changed the first character of to _ instead of A (conforming to what N3, N-Triples, Turtle, etc. use).
 */
/*
RDF_Resource 
newAnonResource(RDF_Parser parser, const char* name) {
  size_t len ;
  char* buff ;
  const char* post = strchr(name, '#');
  if (!post) post = name;
  len = 2 + strlen(parser->url) + 1 + strlen(post) + 1 + numDigits(parser->count) + 1;
  buff = getMem(len);
  snprintf(buff, len, "_:%s-%s-%i", parser->url, post, parser->count++);
  return RDF_GetResource(buff);
}*/

/* JvdG */
/** \todo Check that an XML literal is passed as a string! */
static void RDF_StatementHandler(void *user_data, const raptor_statement *statement)
{
  static const char* rdf_li = "http://www.w3.org/1999/02/22-rdf-syntax-ns#li";
  RDF_Parser rdfp = (RDF_Parser)user_data;
  const char* arc;
  const void* target;
  RDF_ValueType target_type;
  /*  RDF_RESOURCE_TYPE  = 1,
      RDF_INT_TYPE = 2,
      RDF_STRING_TYPE = 3,
      RDF_DATE_TYPE = 4,
      RDF_FLOAT_TYPE = 5,
      RDF_BLOB_TYPE =6,
  */

  if ( statement->subject_type == RAPTOR_IDENTIFIER_TYPE_RESOURCE ) {
    char* src = (char*)statement->subject;
    if ( *src == 'B' && *(src+1) == 'N' ) {
      *(char*)src = '_'; /* This is insanely evil, lets hope it works. */
      *((char*)src+1) = ':';
    }
  }

  switch(statement->predicate_type) {
  case RAPTOR_IDENTIFIER_TYPE_PREDICATE:           /* Predicate URI */
  case RAPTOR_IDENTIFIER_TYPE_ANONYMOUS:           /* _:foo N-Triples, or generated */
    arc = statement->predicate;
    break;
  case RAPTOR_IDENTIFIER_TYPE_ORDINAL:             /* rdf:li, rdf:_<n> etc. */
    arc = rdf_li;
    break;
  case RAPTOR_IDENTIFIER_TYPE_LITERAL:             /* regular literal */
  case RAPTOR_IDENTIFIER_TYPE_XML_LITERAL:         /* rdf:parseType="Literal" */
  case RAPTOR_IDENTIFIER_TYPE_RESOURCE:            /* Resource URI (e.g. rdf:about) */
  case RAPTOR_IDENTIFIER_TYPE_UNKNOWN:             /* Unknown type - illegal */
  default:
    return;
  }

  switch(statement->object_type) {
  case RAPTOR_IDENTIFIER_TYPE_RESOURCE:            /* Resource URI (e.g. rdf:about) */
    target = statement->object;
    target_type = RDF_RESOURCE_TYPE;
    if ( *(char*)target == 'B' && *((char*)target+1) == 'N' ) {
      *(char*)target = '_'; /* This is insanely evil, lets hope it works. */
      *((char*)target+1) = ':';
    }
    break;
  case RAPTOR_IDENTIFIER_TYPE_ANONYMOUS:           /* _:foo N-Triples, or generated */
    target = statement->object;
    target_type = RDF_RESOURCE_TYPE;
    break;
  case RAPTOR_IDENTIFIER_TYPE_LITERAL:             /* regular literal */
    if ( statement->object_literal_datatype ) {
      if ( raptor_uri_equals(statement->object_literal_datatype, xmls_integer) ) {
        char* endchar;
        target = (void*)strtol(statement->object, &endchar, 10);
        target_type = RDF_INT_TYPE;
      }
      else {
        /* Assume string type */
        target = statement->object;
        target_type = RDF_STRING_TYPE;
      }
    }
    else {
      target = statement->object;
      target_type = RDF_STRING_TYPE;
    }
    break;
  case RAPTOR_IDENTIFIER_TYPE_XML_LITERAL:         /* rdf:parseType="Literal" */
    target = statement->object;
    target_type = RDF_STRING_TYPE;
    break;
  case RAPTOR_IDENTIFIER_TYPE_PREDICATE:           /* Predicate URI */
  case RAPTOR_IDENTIFIER_TYPE_ORDINAL:             /* rdf:li, rdf:_<n> etc. */
  case RAPTOR_IDENTIFIER_TYPE_UNKNOWN:             /* Unknown type - illegal */
  default:
    return;
  }

  if ( rdfp->code ) {
    RDF_LoadFileAssert(rdfp, rdfp->db, statement->subject, arc, target, target_type);
  }
  else {
    switch(rdfp->operation) {
    case RDF_ASSERT:
      RDF_Assert(rdfp->db, statement->subject, arc, target, target_type);
      break;
    case RDF_UNASSERT:
      RDF_Unassert(rdfp->db, statement->subject, arc, target, target_type);
      break;
    }
  }

  /* JvdG made this print dots to show progress when loading big files */
  if ( (gAssertionCount & 0xff) == 0xff ) PrintMsg(INFO, ".");
}

/* JvdG
 * Made this use the <lkjlk> syntax for resources.
 */
RDF_ValueType
termType (const char *term) {
  int l = strlen(term);
  int c = 0;
  int isn = 1;

  if (term[0] == '\'') return RDF_STRING_TYPE;
  if (term[0] == '<') return RDF_RESOURCE_TYPE;
  
  while (isn && c < l) {
    if (!isdigit(term[c++])) isn = 0;
  }
  
  if (isn) return RDF_INT_TYPE;
  return 0;
}
      
/* JvdG
 * Made this use the <lkjlk> syntax for resources.
 */
void*
parseTermOfType (char* term, RDF_ValueType type) {
  if (type == RDF_RESOURCE_TYPE) {
    term[strlen(term)-1] = 0;
    return strdup(&term[1]);    
  } else if (type == RDF_INT_TYPE) {
    return (void*)atoi(term);
  } else if (type == RDF_STRING_TYPE) {
    term[strlen(term)-1] = 0;
    return strdup(&term[1]);    
  }
  return 0; /* Shouldn't happen? */
}

void
unloadExistingAssertions(FILE *f, const RDF_DB db, const char* sourceCode) {
  if (f) {
    long pos = ftell(f);
    char *buff = getMem(4096);
    char *line;
    while (0 != (line = fgets(buff, 4096, f))) {
      char *arc, *source, *target;
      arc = line;
      source = strchr(arc, '\t');
      if (source) {
        *source = 0;
        source++;
        target = strchr(source, '\t');
        if (target) {
          const void* value;
          RDF_ValueType type;
          *target = 0;
          target++;
          target[strlen(target)-1] = 0;
          type = termType(target) ;
          value = parseTermOfType(target, type);
          RDF_UnloadFileAssert(sourceCode, db, source, arc, value, type);
        }
      }
    }
    freeMem(buff);
    fseek(f, pos, SEEK_SET);
  }
}

FILE *
rdfpGetFile(RDF_Parser rdfp) {
  return rdfp->file;
}

char *
getFileCode(RDF_Parser rdfp) {
  if (rdfp->code) {
    return rdfp->code;
  } else {
    return 0;
  }
}
