/* -*- Mode: C; tab-width: 4; indent-tabs-mode: nil; c-basic-offset: 2 -*-
 *
 * The contents of this file are subject to the Mozilla Public 
 * License version 1.1 (the "License"); you may not use this file
 * except in compliance with the License. You may obtain a copy of
 * the License at http://www.mozilla.org/MPL/
 *
 * Software distributed under the License is distributed on an "AS
 * IS" basis, WITHOUT WARRANTY OF ANY KIND, either express or 
 * implied. See the License for the specific language governing
 * rights and limitations under the License.
 *
 * The Initial Developer of the Original Code is R.V.Guha. Portions
 * created by R.V.Guha are copyright (c) 2000 R.V.Guha. All Rights
 * Reserved.
 */

/* This is the header file for the functions exported from librdf. 
   Most of these functions have shadow internal functions (rdf-int.h) */

#ifndef RDFDB_LIBRDF_RDF_H
#define RDFDB_LIBRDF_RDF_H

typedef const char*                RDF_Resource;
typedef const char*                RDF_DB;
typedef struct _RDF_FileStruct*    RDF_File;
typedef struct _RDF_CursorStruct*  RDF_Cursor;
typedef struct _RDF_ParserStruct*  RDF_Parser;

/* JvdG
 * Added RDF_VAR_TYPE and RDF_NULL_TYPE.
 */
typedef enum {
  RDF_RESOURCE_TYPE  = 1,
  RDF_INT_TYPE = 2,
  RDF_STRING_TYPE = 3,
  RDF_DATE_TYPE = 4,
  RDF_FLOAT_TYPE = 5,
  RDF_BLOB_TYPE =6,
  /* Not "true" RDF types */
  RDF_NULL_TYPE = 1001,
  RDF_VAR_TYPE = 1000,
} RDF_ValueType;
     
typedef struct _RDF_Value {
  RDF_ValueType type;
  void* value;
} RDF_Value;

typedef enum {
  RDF_OK = 0,
  RDF_UNKNOWN_QUERY_TYPE_ERROR = -1,
  RDF_GENERAL_SYNTAX_ERROR = -2,
  RDF_MALFORMED_LITERAL_ERROR = -3,
  RDF_ERROR = -4,
  RDF_DB_DOES_NOT_EXIST = -5,
  RDF_COULD_NOT_ACCESS_DATA = -6,
  RDF_UNAUTHORIZED_ACCESS = -7,
  RDF_UNCONSTRAINED_VARIABLE = -8,
  RDF_UNKNOWN_FILE_FORMAT_ERROR = -9,
  RDF_WRONG_FILE_PERMISSIONS_ERROR = -10,
  RDF_NO_MORE_VALUES
} RDF_Result;

typedef enum {
  RDF_ASSERT = 1,
  RDF_UNASSERT
} RDF_Operation;

typedef enum {
  RDF_XML,
  RSS,
  DMOZ_RDF_XML_DEPRECATED, /* Not used anymore */
  NTRIPLES,
  TURTLE,
  CSV,
} RDF_SerializationType;



/* Initialize the RDF system. Called when the server initializes. */
RDF_Result RDF_Initialize();

/* Initialize the RDF system. Called when the server initializes. */
RDF_Result RDF_Deinitialize();

/* Get the RDF_Resource corresponding to a URI.  */

const RDF_Resource RDF_GetResource (const char* uri);

/* Get the database corresponding identified by name */

RDF_Result RDF_GetDB (char* name, int createp, RDF_DB *db);

/* Compare two resources for equality */

int RDF_ResourceEq (RDF_Resource r1, RDF_Resource r2);

/* Load the contents of a file into a DB. If the file had already been loaded, the contents are replaced. */

RDF_Result RDF_LoadURLIntoDB (RDF_DB db, const char* url, RDF_SerializationType type,
                              RDF_Operation operation, char** code);

RDF_Result RDF_UnLoadURLFromDB (RDF_DB db, const char* url) ;


/* Assert a triple */

RDF_Result RDF_Assert (RDF_DB db, RDF_Resource source, RDF_Resource arc, const void* target_value, RDF_ValueType target_type) ;

/* Assert a triple as having come from a certain file */

RDF_Result RDF_LoadFileAssert (RDF_Parser rdfp, RDF_DB db, RDF_Resource source, RDF_Resource arc, const void* target_value, RDF_ValueType target_type) ;

/* Unassert a triple */

RDF_Result RDF_Unassert (RDF_DB db, RDF_Resource source, RDF_Resource arc, const void* target_value, RDF_ValueType target_type) ;

/* Unassert a triple as having come from a certain file */

RDF_Result RDF_UnloadFileAssert (const char* sourceCode, RDF_DB db, RDF_Resource source, RDF_Resource arc, const void* target_value, RDF_ValueType target_type);

/* Get a cursor for all the sources with a certain arc  */

RDF_Result RDF_GetArcSources (RDF_DB db, RDF_Resource arc, RDF_Cursor *c) ;

/* Get a cursor for all the targets corresponding to a source.arc pair */

RDF_Result RDF_GetTargets (RDF_DB db, RDF_Resource source, RDF_Resource arc, RDF_Cursor *c) ;

/* Get a cursor for all the source corresponding to a target.arc pair */

RDF_Result RDF_GetSources(RDF_DB db, RDF_Value target, RDF_Resource arc, RDF_Cursor *c) ;

/* Get a cursor for all the arcs going out of a node */

RDF_Result RDF_GetArcsOut(RDF_DB db, RDF_Resource node, RDF_Cursor *c) ;

/* Get a cursor for all the arcs coming into a node */

RDF_Result RDF_GetArcsIn(RDF_DB db, RDF_Value node, RDF_Cursor *c) ;

/* Check whether a particular triple is there in the database */

RDF_Result RDF_HasAssertion(RDF_DB db, const RDF_Resource source, const RDF_Resource arc, const RDF_Value target) ;

/* JvdG Find statements matching the given filter, null values mean "match anything". */

RDF_Result RDF_FindStatements(RDF_DB db, RDF_Resource source, RDF_Resource arc, RDF_Value target,  RDF_Cursor *cursor);

/* Dispose off a cursor */

RDF_Result RDF_DisposeCursor(RDF_Cursor cursor);

/* Advance the cursor */

RDF_Result RDF_Advance (RDF_Cursor c) ;

/* Get the current value of the cursor */

RDF_Result RDF_CursorCurrentValue (RDF_Cursor c, void** value) ;

/* Get the type of the current value of the cursor */

RDF_Result RDF_CursorCurrentValueType (RDF_Cursor c, RDF_ValueType *type) ;

/* Get the size of the current value of the cursor in bytes */

RDF_Result RDF_CursorCurrentValueSize (RDF_Cursor c, int *size);

/* JvdG Get the current source (only valid with RDF_FindStatements). */

RDF_Result RDF_CursorCurrentSource (RDF_Cursor c, void** value);

/* JvdG Get the current arc (only valid with RDF_FindStatements). */

RDF_Result RDF_CursorCurrentArc (RDF_Cursor c, void** value);

/* JvdG Export the contents of a database to a file. */

RDF_Result RDF_ExportDB(RDF_DB db, const char* url, RDF_SerializationType type, int raw);

#endif/*RDFDB_LIBRDF_RDF_H*/
