/* -*- Mode: C; tab-width: 4; indent-tabs-mode: nil; c-basic-offset: 2 -*-
 *
 * The contents of this file are subject to the Mozilla Public 
 * License version 1.1 (the "License"); you may not use this file
 * except in compliance with the License. You may obtain a copy of
 * the License at http://www.mozilla.org/MPL/
 *
 * Software distributed under the License is distributed on an "AS
 * IS" basis, WITHOUT WARRANTY OF ANY KIND, either express or 
 * implied. See the License for the specific language governing
 * rights and limitations under the License.
 *
 * The Initial Developer of the Original Code is R.V.Guha. Portions
 * created by R.V.Guha are copyright (c) 2000 R.V.Guha. All Rights
 * Reserved.
 *
 * Contributors : Eric Van Der Vlist, Ivan Kurmanov, Jasper van de Gronde (win32 port)
 */

/** \file */

#include "../utils/utils.h"
#include "queryParser.h"
#include "selectParser.h"
#include <regexp/regexp.h>
#include <stdio.h>
#ifdef WIN32
#include <time.h>
#endif

extern int       DB_Flush (const char* db) ;
extern int gAssertionCount;
void addNamespace (ClientConnection cl, char *prefix, char* url);
void removeNamespace (ClientConnection cl, char* prefix);

static void freeQuery (Query nq);

#define VALUE_BUF 4048

#define DEBUG_PROCESS_QUERY 0

/* Jasper van de Gronde
 * Holds the global namespace mappings.
 */
NamespaceStruct* gNamespaceList = 0;

/* Jasper van de Gronde
 * This initializes anything processQuery might need.
 * At the moment it just initializes the global namespace list.
 */
int
initProcessQuery() {
  const char* const namespaces = get_setting("NAMESPACES");
  if ( namespaces ) {
    NamespaceStruct* curnamespace = gNamespaceList;
    const char* curprefixbegin = 0;
    const char* curprefixend = 0;
    const char* curnsbegin = 0;
    const char* curpos;
    for(curpos = namespaces; *curpos; ++curpos) {
      if ( !curprefixbegin ) {
        curprefixbegin = curpos;
      }
      else if ( !curprefixend ) {
        if ( *curpos == '<' ) {
          curprefixend = curpos;
          curnsbegin = curpos+1;
        }
      }
      else {
        if ( *curpos == '>' ) {
          if ( !curnamespace ) {
            gNamespaceList = (NamespaceStruct*)getMem(sizeof(NamespaceStruct));
            curnamespace = gNamespaceList;
          }
          else {
            curnamespace->next = (NamespaceStruct*)getMem(sizeof(NamespaceStruct));
            curnamespace = curnamespace->next;
          }
          if ( !curnamespace ) break;
          curnamespace->prefix = strduprange(curprefixbegin, curprefixend);
          curnamespace->url = strduprange(curnsbegin, curpos);

          curprefixbegin = 0;
          curprefixend = 0;
          curnsbegin = 0;
        }
      }
    }
  }
  return 0;
}

/* Jasper van de Gronde
 * This deinitializes anything InitializeProcessQuery might have initialized.
 */
int
deinitProcessQuery() {
  NamespaceStruct* next;
  while(gNamespaceList) {
    if ( gNamespaceList->prefix ) freeMem(gNamespaceList->prefix);
    if ( gNamespaceList->url ) freeMem(gNamespaceList->url);
    next = gNamespaceList->next;
    freeMem((char*)gNamespaceList);
    gNamespaceList = next;
  }
  return 0;
}

/* JvdG
 * Made this accept an RDF_Value
 * and changed v->value == 0 to v->value.type == 0.
 * Then changed it to just accept a Variable.
 */
static
int
varp (const Variable v) {
  return v && v->value.type == 0;
}

/* JvdG
 * Made this accept and return an RDF_Value*.
 * Added a Variable as parameter.
 */
static
const RDF_Value*
vval (const RDF_Value* var, const Variable v) {
  if ( v && var->type == RDF_VAR_TYPE ) {
    return &(v->value);
  }
  else {
    return var;
  }
}

/* JvdG
 */
static
Variable
searchVariable  (Query q, const RDF_Value val) {
  if ( val.type == RDF_VAR_TYPE ) {
    const char* const var = val.value;
    Variable v = q->variables;
    while (v) {
      if (streq(v->name, var)) {
        return v;
      }
      v = v->next;
    }
  }
  return 0;
}

/* JvdG
 * Made this accept an RDF_Value.
 * Changed it to accept a Variable.
 */
static
void
setVariableValue  (Variable var, RDF_Value value) {
  var->value = value;
}
   
/* JvdG
 */
static
int
isVisibleVariable  (Variable var) {
  return var->outputVariablep;
}


/** Jasper van de Gronde
 * This checks wether the var-value-type combination isn't in conflict with the constraint list.
 * \todo This could be faster if the constraints are sorted (on variable name), optionally
 *       with a binary search (which would involve a change in the way constraints are
 *       stored), but it would probably be faster (and easier) to simply stop the loop as
 *       soon as strcmp(var, constraint->expr.var) < 0, at least for small numbers of
 *       constraints.
 */
static
BOOL
checkConstraintList(Query q, Variable var, RDF_Value value) {
  struct constraint_list_t* constraint;

  /* this is done because the old code left the '?' in front of the variable name,
     while the new code doesn't (so ?x in the old code is just x in the new code) */
  /* if ( var && *var ) var += 1; */

  constraint = q->constraint_list;
  while(constraint) {
    if ( streq(var->name, constraint->expr.var) ) {
      if ( constraint->expr.var_or_literal.type != value.type ) return FALSE;
      switch(constraint->expr.operation) {
      /* NOTE: These all test for the opposite, as they return false if it doesn't match */
      /* Replaced STRCOMP_LIKE and STRCOMP_SOMETHING by STRCOMP_MATCH and STRCOMP_NMATCH, following http://www.w3.org/Submission/2004/SUBM-RDQL-20040109/ */
      /* According to RDQL these two should be case insensitive, see http://www.w3.org/Submission/2004/SUBM-RDQL-20040109/ */
      case STRCOMP_NE:
        if ( 0 == strcasecmp((char*)value.value, constraint->expr.var_or_literal.text) ) return FALSE;
        break;
      case STRCOMP_EQ:
        if ( 0 != strcasecmp((char*)value.value, constraint->expr.var_or_literal.text) ) return FALSE;
        break;
      case STRCOMP_MATCH:
        if ( !regexec(constraint->expr.rx, (char*)value.value) ) return FALSE;
        break;
      case STRCOMP_NMATCH:
        if ( regexec(constraint->expr.rx, (char*)value.value) ) return FALSE;
        break;

      case NUMCOMP_GT: /* '>' */
        if ( (int)value.value <= constraint->expr.var_or_literal.num ) return FALSE;
        break;
      case NUMCOMP_LT: /* '<' */
        if ( (int)value.value >= constraint->expr.var_or_literal.num ) return FALSE;
        break;
      case NUMCOMP_EQ: /* both '==' and '=' */
        if ( (int)value.value != constraint->expr.var_or_literal.num ) return FALSE;
        break;
      case NUMCOMP_NE: /* both '!=' and '<>' */
        if ( (int)value.value == constraint->expr.var_or_literal.num ) return FALSE;
        break;
      case NUMCOMP_LTEQ: /* '<=' */
        if ( (int)value.value > constraint->expr.var_or_literal.num ) return FALSE;
        break;
      case NUMCOMP_GTEQ: /* '>=' */
        if ( (int)value.value < constraint->expr.var_or_literal.num ) return FALSE;
        break;

      default:
        /* Shouldn't occur as this should have triggered a syntax error. */
        return FALSE;
      }
    }
    constraint = constraint->next;
  }
  return TRUE;
}

/* JvdG
 * Called from processQueryInt. This way it always appears and it always appears exactly once.
 */
static
void
printHeader (Query q) {
  Variable v = q->variables;
  if (q->outputType == TAB_LIMITED_OUTPUT) {
    while (v) {
      if (v->outputVariablep) {
        socket_stream_write(q->cl->socket_stream, "?", strlen("?"));
        socket_stream_write(q->cl->socket_stream, v->name, strlen(v->name));
        socket_stream_write(q->cl->socket_stream, "\t", 1);
      }
      v = v->next;
    }
    socket_stream_write(q->cl->socket_stream, NETWORK_NEWLINE, NETWORK_NEWLINE_LEN);
  }
}

/* JvdG
 * Modified this to work with RDF_Value's.
 * I also added some checks.
 * Removed the header printing bit, see the printHeader function above.
 */
static
int
addAnswer (Query q) {
  Variable v = q->variables;

  while (v) {
    if (v->outputVariablep) {
      char ansLine[VALUE_BUF];

      if (v->value.type == RDF_RESOURCE_TYPE) {
        if( !v->value.value || strlen( (char*)v->value.value ) > VALUE_BUF-30 ) { 
          /*** this is an exception and is a good reason to complain and die ***/
        }
        if (q->outputType == TAB_LIMITED_OUTPUT) {
          snprintf(ansLine, VALUE_BUF, "<%s>\t", (char*)v->value.value);
        } else {
          snprintf(ansLine, VALUE_BUF, "?%s = <%s>\t", v->name, (char*)v->value.value);
        }
      }
      else if (v->value.type == RDF_STRING_TYPE) {
        char* destination = ansLine;

        if( !v->value.value || strlen( (char*)v->value.value ) > VALUE_BUF-30 ) { 
          /*** this is an exception and is a good reason to complain and die ***/
        }

        /**  value start  **/ 
        if (q->outputType == TAB_LIMITED_OUTPUT) {
        } else { 
          snprintf( destination, VALUE_BUF, "?%s = ", v->name );
          destination += strlen( destination );
        }

        /**  here goes the value itself  **/
        destination = outputEscapedString((char*)v->value.value, destination);

        (*destination++) = '\t';
        (*destination++) = 0;
        
      }
      else if (v->value.type == RDF_INT_TYPE) {
        if (q->outputType == TAB_LIMITED_OUTPUT) 
          snprintf(ansLine, VALUE_BUF, "%i\t", (int)v->value.value);
        else 
          snprintf(ansLine, VALUE_BUF, "?%s = %i\t", v->name, (int)v->value.value);
      }
      else if (v->value.type == RDF_NULL_TYPE) {
        strcpy(ansLine, "\t");
      }
      else {
        snprintf(ansLine, VALUE_BUF, "UNKOWNTYPE(%i)\t", (int)v->value.type);
      }
      socket_stream_write(q->cl->socket_stream, ansLine, strlen(ansLine));
      FastPrintMsg1(DEBUG_INFO, "%s", ansLine);
    }
    v = v->next;
  }
  FastPrintMsg0(DEBUG_INFO, "\n");
  if ( socket_stream_write(q->cl->socket_stream, NETWORK_NEWLINE, NETWORK_NEWLINE_LEN) < 0 ) {
    PrintMsg(FAILURE, "Client closed the connection!\n");
    return 0;
  }
  return 1;
}


/**  
     recursive Select function:
       int SelectNextInt 
     parameters: 
       Query q, 
       triple Literal nl 
     returns:
       0 upon normal finish;
       !0 unon error (error code);
     
**/
static
int
SelectNextInt (Query q, Statement* const nl) {
  if (q->abort) return 0;

  if (!nl) {

    if( q->outputType > 0 && q->non_null ) {
      /* we must print out what we found */ 
      if (!addAnswer(q)) {
        q->abort = 1;
        return 0;
      }
    }

    /*    DEBUG printf ( "-query outputType: %i\n", q->outputType ); */
    /*    DEBUG printf ( "-after procesing a valid assertion\n" ); */

  } else {
    int err;
    /* JvdG added these variables for more clarity and perhaps performance (depending on what the compiler's optimizer does). */
    const Variable source_var = searchVariable(q, nl->source);
    const Variable arc_var = searchVariable(q, nl->arc);
    const Variable target_var = searchVariable(q, nl->target);
    const RDF_Value* source_vval = vval(&(nl->source), source_var);
    const RDF_Value* arc_vval = vval(&(nl->arc), arc_var);
    const RDF_Value* target_vval = vval(&(nl->target), target_var);
    const int source_is_var = varp(source_var); /*(source_var ? source_var->value.type == 0 : 0);/* varp(q, nl->source); */
    const int arc_is_var = varp(arc_var); /*(arc_var ? arc_var->value.type == 0 : 0);/* varp(q, nl->arc); */
    const int target_is_var = varp(target_var); /*(target_var ? target_var->value.type == 0 : 0);/* varp(q, nl->target); */
    const int has_null = (
      (!source_is_var && source_vval->type == RDF_NULL_TYPE) ||
      (!arc_is_var && arc_vval->type == RDF_NULL_TYPE) ||
      (!target_is_var && target_vval->type == RDF_NULL_TYPE)
      );

    if (!source_is_var && !arc_is_var && !target_is_var) {
      /* 
         that means: if all three elements of the triple are already exact values, 
         either specified by the user in the original query, 
         or found as satisfying to the original query condition 
      */

      err = 0; /* No error if has_null is true */
    
      /* 
         just checking ... and recursing to check that if would satisfy other 
         conditions as well 
      */
      if (has_null || 0 != (err = RDF_HasAssertion(q->database, source_vval->value, 
                                        arc_vval->value, 
                                       *target_vval)))
      {
        if (err < 0) return err;

        SelectNextInt(q, nl->nxt);
      }

    } else {  /*  at least one of the triple elements is a variable  */ 

      /***  
            therefore we need to search for data triples, that satisfy
            the condition requirement.  We do this by calling
            RDF_Get... (..., &c ) function, which would get us the data
            in the RDF_Cursor c;
          
            Also, if in the current triple there are two parameter variables, then we will 
            have to cycle through all the first variable's values' first, looking for possible
            second variable values at the same time....
      ***/

      RDF_Cursor c = 0;
      Statement* next;
      int len;
      Variable var; /* JvdG Changed this from a const char* to a Variable */
      BOOL matchedSomething = FALSE;

      /* JvdG Moved this here, it used to be way below, just before the while(c). */
      if ( (nl->source.type == RDF_VAR_TYPE && !source_var)
        || (nl->arc.type == RDF_VAR_TYPE && !arc_var)
        || (nl->target.type == RDF_VAR_TYPE && !target_var) ) {
        PrintMsg(FAILURE, "\nvariable not found!\n");
        return RDF_ERROR;
      }

      /* JvdG Added this check.
         A vval can't be null, as it either points to the value field of a non-null Variable,
         or to the address of the source, arc or target field of a non-null Literal. */
      if ( (!source_is_var && source_vval->type != RDF_RESOURCE_TYPE)
        || (!arc_is_var && arc_vval->type != RDF_RESOURCE_TYPE) ) {
        /* The source and/or arc is not a URI (a literal most likely), if this is the case
           we simply don't match anything. It's no good trying to return an error here, as
           we should have reached this situation due to finding a match for a variable
           (which is perfectly legal). And as it is possible that this is an optional triple
           we simply don't match anything and let the normal logic do its work. */
        int nvars = 0;
        var = 0;
        if ( source_is_var ) {
          ++nvars;
          if ( !var ) var = source_var;
        }
        if ( target_is_var ) {
          ++nvars;
          if ( !var ) var = target_var;
        }
        if ( arc_is_var ) {
          ++nvars;
          if ( !var ) var = arc_var;
        }
        if ( nvars == 1 ) {
          /* We know for sure nvars == 2 here, as due to the condition of the if about 80
             lines above this one at least one of the values is a var. */
          next = nl->nxt; /* This triple won't match anything, no matter what the other values
                             are, so we simply move to the next. */
        }
        else {
          /* We know for sure nvars == 2 here, as due to the condition of the if above at
             least either source or arc is not a var. */
          next = nl;
        }
      }
      else {
        /* A perfectly valid triple :) */
        if (source_is_var && !arc_is_var && !target_is_var) {
          /*** source (resource) is a variable ***/
          var = source_var;
          next = nl->nxt;
          if ( has_null ) {
            err = 0;
          }
          else {
            err = RDF_GetSources(q->database, 
                                 *target_vval, 
                                 arc_vval->value, &c);
          }
        }
        else if (!source_is_var && !arc_is_var && target_is_var) {
          /*** target (object) is a variable ***/
          var = target_var;
          next = nl->nxt;
          if ( has_null ) {
            err = 0;
          }
          else {
            err = RDF_GetTargets(q->database, 
                                 source_vval->value, 
                                 arc_vval->value,  &c);
          }
        }
        else if (source_is_var && !arc_is_var && target_is_var) {
          /*** source (resource) and target are variables ***/
          var = source_var;
          next = nl;
          if ( has_null ) {
            err = 0;
          }
          else {
            err = RDF_GetArcSources(q->database, arc_vval->value,  &c);
          }
        }
        else if (!source_is_var && arc_is_var) {
          /***  arc (property) is a variable, but the source (subject) is not  ***/
          var = arc_var;
          next = nl; /* (target_is_var ? nl : nl->nxt); Won't work because we need to check that the triple actually exists anyway. */
          if ( has_null ) {
            err = 0;
          }
          else {
            err = RDF_GetArcsOut(q->database, source_vval->value,  &c);
          }
        }
        else if (arc_is_var && !target_is_var) {
          /***  arc (property) is a variable, but the target (object) is not  ***/
          var = arc_var;
          next = nl; /* (source_is_var ? nl : nl->nxt); Won't work because we need to check that the triple actually exists anyway. */
          if ( has_null ) {
            err = 0;
          }
          else {
            err = RDF_GetArcsIn(q->database, *target_vval,  &c);
          }
        }
        else {
          /*  all three triple elements are unconstrained variables   */
          /* JvdG Wrote this body to add support for (?a ?b ?c), used to return RDF_ERROR here. */
          const RDF_Value RDFNull = {0,0};
          var = 0; /* We write into all three at once. */
          next = nl->next;
          err = RDF_FindStatements(q->database, 0, 0, RDFNull, &c);
          PrintMsg(INFO, "Returning ALL triples in the DB, this is slow and should be avoided.\n");
        }

#if DEBUG_PROCESS_QUERY
        PrintMsg(DEBUG_INFO, "\n-starting cursor: %x\n", c );
#endif

        if (err) return err;
        if (!c && !has_null) return RDF_ERROR;
    
        if (c) {
          int old_non_null = q->non_null; /* JvdG Added to make it easier to restore this value. */
          while (0 == RDF_Advance(c)) {
            RDF_Value cur_source;
            RDF_Value cur_arc;
            RDF_Value value;

#if DEBUG_PROCESS_QUERY
            PrintMsg(DEBUG_INFO, "-(another) item found in cursor %x\n", c );
#endif

            RDF_CursorCurrentValue(c, &(value.value));
            RDF_CursorCurrentValueSize(c, &len);
            RDF_CursorCurrentValueType(c, &(value.type));
            if (value.type == RDF_RESOURCE_TYPE || value.type == RDF_STRING_TYPE) {
              ((char*)value.value)[len] = 0; 
            }

            if ( !var ) {
              /* JvdG We're reading whole triples! */
              RDF_CursorCurrentSource(c, &(cur_source.value));
              RDF_CursorCurrentArc(c, &(cur_arc.value));
              cur_source.type = RDF_RESOURCE_TYPE;
              cur_arc.type = RDF_RESOURCE_TYPE;
            }

#if DEBUG_PROCESS_QUERY
            if ( var ) {
              PrintMsg(DEBUG_INFO, "-variable %s's value is '%s'\n", var, value.value );
            }
            else {
              /* JvdG We're reading whole triples! */
              PrintMsg(DEBUG_INFO, "-variable %s's value is '%s'\n", var, value.value );
            }
#endif

            /* Jasper van de Gronde
             * Added the checkConstraint bit
             */
            if ( ( var ? checkConstraintList(q, var, value)
              : (checkConstraintList(q, source_var, cur_source) && checkConstraintList(q, arc_var, cur_arc) && checkConstraintList(q, target_var, value))
              ) ) {
              if ( var ) {
                /* JvdG Just one var at a time. */
                setVariableValue(var, value);
              }
              else {
                /* JvdG We're reading whole triples! */
                setVariableValue(source_var, cur_source);
                setVariableValue(arc_var, cur_arc);
                setVariableValue(target_var, value);
              }

              if ( !matchedSomething ) {
                matchedSomething = TRUE;
                if ( var ) {
                  if ( isVisibleVariable(var) ) ++q->non_null;
                }
                else {
                  if ( isVisibleVariable(source_var) ) ++q->non_null;
                  if ( isVisibleVariable(arc_var) ) ++q->non_null;
                  if ( isVisibleVariable(target_var) ) ++q->non_null;
                }
              }

              /**** delete if necessary ****/
              if( (q->type == DELETEWHERE) && (!next) ) {
                char * src, *arc;
                const RDF_Value* target;
                int err;

#if DEBUG_PROCESS_QUERY
                PrintMsg(DEBUG_INFO, "-deleting an assertion\n" );
#endif

                src    = (char*)source_vval->value;/*(vval( q, &(nl->source) )->value);*/
                arc    = (char*)arc_vval->value;/*(vval( q, &(nl->arc)    )->value);*/
                target =        target_vval;/*vval( q, &(nl->target) );*/
        
#if DEBUG_PROCESS_QUERY
                PrintMsg(DEBUG_INFO, "'%s' = %s => '%s' (type: %i)\n", 
                               src, arc, target->value, target->type
                               ) ;
#endif
        
                err = RDF_Unassert( q->database, 
                                    (const RDF_Resource) src, 
                                    (const RDF_Resource) arc, 
                                    target->value, target->type) ;
        
                /*        if (err < 0) return err; */
              }

              err = SelectNextInt(q, next);

              if (err) {
                RDF_DisposeCursor(c);
                q->non_null = old_non_null;
                return err;
              }
            } /* if ( checkConstraintList(q, var, value) ) */

          } /* while(0 == RDF_Advance(c)) */
          RDF_DisposeCursor(c);
          q->non_null = old_non_null;
        } /* if (c) */
#if DEBUG_PROCESS_QUERY
        PrintMsg(DEBUG_INFO, "-no more items in cursor\n" );
#endif
      } /* if ( check that source and arc are either a var or a URI ) */

      if ( !matchedSomething && nl->optional ) {
        const RDF_Value value = {RDF_NULL_TYPE, 0};
        if ( ( var ? checkConstraintList(q, var, value)
          : (checkConstraintList(q, source_var, value) && checkConstraintList(q, arc_var, value) && checkConstraintList(q, target_var, value))
          ) ) {
          if ( var ) {
            /* JvdG Just one var at a time. */
            setVariableValue(var, value);
          }
          else {
            /* JvdG We're reading whole triples! */
            setVariableValue(source_var, value);
            setVariableValue(arc_var, value);
            setVariableValue(target_var, value);
          }
          err = SelectNextInt(q, next);
          if (err) return err;
        }
      }

      {
        RDF_Value value = {0, 0};
        if ( var ) {
          /* JvdG Just one var at a time. */
          setVariableValue(var, value);
        }
        else {
          /* JvdG We're reading whole triples! */
          setVariableValue(source_var, value);
          setVariableValue(arc_var, value);
          setVariableValue(target_var, value);
        }
      }
    }
  }

#if DEBUG_PROCESS_QUERY
  PrintMsg(DEBUG_INFO, "-quitting SelectNextInt normally\n" );
#endif

  return 0;
}

/* Jasper van de Gronde
 * Changed this to not put anything in between the prefix and the namespace.
 */
static
char*
appendNamespace (const char* str, Namespace ns) {
 char* st =  strchr(str, ':');
 if (!st) {
   if (strlen(ns->prefix) > 0) {
     return strdup(str);
   } else {
     return concat(ns->url, str);
   }
 } else {
   return concat(ns->url, st+1);
 }
}

static
int
namespaceMatch(const char* uri, Namespace ns) {
  if (strlen(ns->prefix) == 0) {
    return (strchr(uri, ':') == 0);
  } else {
    const char* st = strchr(uri, ':');
    if (!st) return 0;
    if ((size_t)(st-uri) != strlen(ns->prefix)) return 0;
    return (strncmp(uri, ns->prefix, st-uri) == 0);
  }
}

/* JvdG */
static
void
namespaceMungeRDFValue (Namespace ns, RDF_Value* v) {
  if (v->type == RDF_RESOURCE_TYPE) {
    Namespace nsx = ns;
    while (nsx) {
      if (namespaceMatch(v->value, nsx)) {
        const char *nn = appendNamespace(v->value, nsx);
        freeMem(v->value);
        v->value = (void*)nn;
        break;
      }
      nsx = nsx->next;
    }
  }
}

/* JvdG Made this use namespaceMungeRDFValue */
static
void
namespaceMungeLiteral (Namespace ns, Statement* l) {
  namespaceMungeRDFValue(ns, &(l->source));
  namespaceMungeRDFValue(ns, &(l->arc));
  namespaceMungeRDFValue(ns, &(l->target));
}


/* Jasper van de Gronde
 * Does practically the same as namespaceMungeLiteral, only for a constraint.
 */
static
void
namespaceMungeConstraint (Namespace ns, struct constraint_list_t* l) {
  if (l->expr.var_or_literal.type == RDF_RESOURCE_TYPE) {
    char* resource = l->expr.var_or_literal.text;
    Namespace nsx = ns;
    while (nsx) {
      if (namespaceMatch(resource, nsx)) {
        char *nn = appendNamespace(resource, nsx);
        freeMem(l->expr.var_or_literal.text);
        l->expr.var_or_literal.text = nn;
        break;
      }
      nsx = nsx->next;
    }
  }
}

static
Statement*
orderStatements1 (Query q) {
  Statement* l;
  Statement* n;
  n = 0;
  for (l = q->statements; l != 0; l = l->next) {
    l->score = 0;
    if (varp(searchVariable(q, l->arc))) l->score++;
    if (varp(searchVariable(q, l->source))) l->score++;
    if (varp(searchVariable(q, l->target))) l->score++;
    if (l->optional) l->score += 10*l->optional; /* JvdG This makes sure optional triples are always evaluated after non-optional triples. */
  }
  for (l = q->statements; l != 0; l = l->next) {
    if (!l->taken) {
      n = l; break;
    }
  }
  for (l = q->statements; l != 0; l = l->next) {
    if (!l->taken && l->score < n->score) 
      n = l;
  }
  return n;
}

/**** 
      reorder Statements function: void orderStatements() 

      the triples of the query's where statement are reordered by this
      function and some preparational processing is being done at this
      level

      new order of the triples is saved in the statement->nxt, instead
      of statement->next original by the queryParser
*****/

static
void
orderStatements (Query q) {
  Statement* next;
  Statement* prev;
  ASSERT(q->stats == 0);
  for (next =  orderStatements1(q); next != 0; next = orderStatements1(q)) {
    const Variable source_var = searchVariable(q, next->source);
    const Variable target_var = searchVariable(q, next->target);
    if (!q->stats) {
      q->stats = next;
    } else {
      prev->nxt = next;
    }
    if (varp(source_var)) {
      const RDF_Value tempval = {RDF_STRING_TYPE, "temp"};
      setVariableValue(source_var, tempval);
    }
    if (varp(target_var)) {
      const RDF_Value tempval = {RDF_STRING_TYPE, "temp"};
      setVariableValue(target_var, tempval);
    }
    next->taken = 1;
    prev = next;
  }
  {
    Variable v ;
    for (v = q->variables; v != 0; v = v->next) {
      v->value.type = 0;
      v->value.value = 0;
    }
  }
}


static
void
describeQuery (Query q) {
  int n = 1;
  Statement* stat = q->statements;
  Variable var = q->variables;
  PrintMsg(FORCE, "Optype = %i\nDB=%s\n", q->type, q->database);
  if (stat) {
    PrintMsg(FORCE, "Literals:\n");
    while (stat) {
      if ( stat->target.type != RDF_INT_TYPE ) {
        PrintMsg(FORCE, "%i) source=%s, arc=%s, target=%s\n", n++, stat->source.value, stat->arc.value, stat->target.value);
      }
      else {
        PrintMsg(FORCE, "%i) source=%s, arc=%s, target=%i\n", n++, stat->source.value, stat->arc.value, (int)(stat->target.value));
      }
      stat = stat->next;
    }
  }
  if (var) {
    n=1;
    PrintMsg(FORCE, "Variables:\n");
    while (var) {
      PrintMsg(FORCE, "%i) name=?%s\n", n++, var->name);
      var = var->next;
    }
  }
}

static
int
processQueryInt (const char* queryString, Query *qpt, ClientConnection cl) {
  QueryParseResult err;
  Query            query;
  Statement* l ;
  struct constraint_list_t* c;

  if (IsLevelEnabled(FAILURE|DEBUG_INFO)) {
    time_t tm ;
    tm = time(&tm);
    PrintMsg(INFO, "%sQuery> %s\n", ctime(&tm), queryString);
  }


  if (0 != (err = parseQuery(queryString, qpt))) {
    return err;
  }


  query = *qpt;

  query->cl = cl;

  /* Jasper van de Gronde
   * Added the namespaceMungeConstraint and gNamespaceList bits.
   */
  if (query->for_list) {
    for (l = query->statements; l != 0; l = l->next) {
      namespaceMungeLiteral(query->for_list, l);
    }
    for (c = query->constraint_list; c != 0; c = c->next) {
      namespaceMungeConstraint(query->for_list, c);
    }
  }
  if (cl->namespaces) {
    for (l = query->statements; l != 0; l = l->next) {
      namespaceMungeLiteral(cl->namespaces, l);
    }
    for (c = query->constraint_list; c != 0; c = c->next) {
      namespaceMungeConstraint(cl->namespaces, c);
    }
  }
  if (gNamespaceList) {
    for (l = query->statements; l != 0; l = l->next) {
      namespaceMungeLiteral(gNamespaceList, l);
    }
    for (c = query->constraint_list; c != 0; c = c->next) {
      namespaceMungeConstraint(gNamespaceList, c);
    }
  }


  if (IsLevelEnabled(FAILURE|DEBUG_INFO)) {
    describeQuery (query) ;
  }

  switch (query->type) {

  case DELETEWHERE :    
  case SELECT :
    {
      if ( !action_allowed(cl->security_context, action_read_db | (query->type == SELECT ? 0 : action_write_db), 0) ) {
        err = RDF_UNAUTHORIZED_ACCESS;
      }
      else {
        orderStatements(query);
        printHeader(query);
        if (0 != (err = SelectNextInt(query, query->stats))) {
          return err;
        }
      }
      break;
    }

  case INSERTFILE:
  case DELETEFILE:
  case UPLOAD :
    {
      if ( !action_allowed(cl->security_context, action_write_db | action_read_url, 0) ) {
        err = RDF_UNAUTHORIZED_ACCESS;
      }
      else {
        char* code = 0;
        int start = gAssertionCount;
        switch(query->type) {
        case INSERTFILE:
          /* File insert query */
          err = RDF_LoadURLIntoDB (query->database, query->url, query->fileType, RDF_ASSERT, 0);
          break;
        case DELETEFILE:
          /* File delete query */
          err = RDF_LoadURLIntoDB (query->database, query->url, query->fileType, RDF_UNASSERT, 0);
          break;
        case UPLOAD:
          /* Load query */
          err = RDF_LoadURLIntoDB (query->database, query->url, query->fileType, RDF_ASSERT, &code);
          break;
        }
        DB_Flush (query->database) ;
        switch(query->type) {
        case INSERTFILE:
          /* File insert query */
          PrintMsg(INFO, "\n%i Assertions added\n", gAssertionCount-start);
          snprintf(query->error_message, QUERY_ERR_LEN, "%i:%i Assertions added", gAssertionCount-start, gAssertionCount-start);
          break;
        case DELETEFILE:
          /* File delete query */
          PrintMsg(INFO, "\n%i Assertions removed\n", gAssertionCount-start);
          snprintf(query->error_message, QUERY_ERR_LEN, "%i:%i Assertions removed", gAssertionCount-start, gAssertionCount-start);
          break;
        case UPLOAD:
          /* Load query */
          PrintMsg(INFO, "\n%i Assertions added\n", gAssertionCount-start);
          snprintf(query->error_message, QUERY_ERR_LEN, "%i:%i Assertions added", gAssertionCount-start, gAssertionCount-start);
          break;
        }
        if (code) {
          /* JvdG Queries which have no result should not return any results, therefor I
                  commented out the following (and put the information in the error message):
          writen(cl->socket, code, strlen(code));
          writen(cl->socket, NETWORK_NEWLINE, NETWORK_NEWLINE_LEN);
          */
          snprintf(query->error_message, QUERY_ERR_LEN, "%s", code);
        }
      }
      break;
    }

  case UNLOAD :
    {
      if ( !action_allowed(cl->security_context, action_write_db | action_read_url, 0) ) {
        err = RDF_UNAUTHORIZED_ACCESS;
      }
      else {
        int start = gAssertionCount;
        err = RDF_UnLoadURLFromDB (query->database, query->url);
        DB_Flush (query->database) ;
        PrintMsg(INFO, "\n%i Assertions removed\n", gAssertionCount-start);
        snprintf(query->error_message, QUERY_ERR_LEN, "%i:%i Assertions removed", gAssertionCount-start, gAssertionCount-start);
      }
      break;
    }

  case EXPORT :
    {
      if ( !action_allowed(cl->security_context, action_read_db | action_write_url, 0) ) {
        err = RDF_UNAUTHORIZED_ACCESS;
      }
      else {
        int start = gAssertionCount;
        err = RDF_ExportDB(query->database, query->url, query->fileType, query->exportRaw);
        PrintMsg(INFO, "\n%i Assertions exported\n", gAssertionCount-start);
        snprintf(query->error_message, QUERY_ERR_LEN, "%i:%i Assertions exported", gAssertionCount-start, gAssertionCount-start);
      }
      break;
    }

  case CREATE :
    {
      if ( !action_allowed(cl->security_context, action_admin, 0) ) {
        err = RDF_UNAUTHORIZED_ACCESS;
      }
      else {
        RDF_DB db;
        if ( (err=RDF_GetDB(query->database, 0, &db)) == RDF_DB_DOES_NOT_EXIST ) {
          err = RDF_GetDB(query->database, 1, &db);
        }
        else if ( err == 0 ) {
          err = RDF_ERROR;
          snprintf(query->error_message, QUERY_ERR_LEN, ":Database already exists.");
        }
      }
      break;
    }
    
  case INSERT :
    {
      if ( !action_allowed(cl->security_context, action_write_db, 0) ) {
        err = RDF_UNAUTHORIZED_ACCESS;
      }
      else {
        /* Normal insert query: insert into DB TRIPLES */
        Statement* l = query->statements;
        while (l) {
          err = RDF_Assert(query->database, l->source.value, l->arc.value, l->target.value, l->target.type);
          l = l->next;
        }
        DB_Flush (query->database);
      }
      break;
    }

  case DELETE :    
    {
      if ( !action_allowed(cl->security_context, action_write_db, 0) ) {
        err = RDF_UNAUTHORIZED_ACCESS;
      }
      else {
        Statement* l = query->statements;
        while (l) {
          err = RDF_Unassert(query->database, l->source.value, l->arc.value, l->target.value, l->target.type);
          l = l->next;
        }
        DB_Flush (query->database);
      }
      break;
    }

  case ENTER :    
    {
      char* prefix = query->namespacePrefix;
      if (startsWith(prefix, "xmlns:")) {
        addNamespace(cl, &prefix[6], query->namespace);
      } else {
        addNamespace(cl, "", query->namespace);
      }
      break;
    }

  case LEAVE :    
    {
      char* prefix = query->namespacePrefix;
      if (startsWith(prefix, "xmlns:")) {
        removeNamespace(cl, &prefix[6]/*, query->namespace*/);
      } else {
        removeNamespace(cl, ""/*, query->namespace*/);
      }
      break;
    }

  case LOGIN:
    {
      if ( authenticate(cl->security_context, &query->identity) ) {
        err = RDF_OK;
      }
      else {
        err = RDF_ERROR;
        snprintf(query->error_message, QUERY_ERR_LEN, ":Invalid user/password pair.");
      }
      break;
    }

  }
  
  /*   write(fd, "%i", q->id); */
  if (err) {
    return err;
  } else {
    return 0; /* JvdG Not used anymore: query->id;*/
  }
}

#define HELP_MSG "Please look at http://web1.guha.com/rdfdb/ for help."
/* JvdG Added extra help message. */
#define NEW_HELP_MSG "See http://home.hccnet.nl/th.v.d.gronde/ for information on the Win32 port and the features of this modification."

#define errStrLen 32+QUERY_ERR_LEN

int
ProcessQuery (const char* query, ClientConnection cl) {
  char errStr[errStrLen+1];

  if (streq(query, "help")) {
    socket_stream_write(cl->socket_stream, HELP_MSG, strlen(HELP_MSG));
    socket_stream_write(cl->socket_stream, NETWORK_NEWLINE, strlen(NETWORK_NEWLINE));
    socket_stream_write(cl->socket_stream, NEW_HELP_MSG, strlen(NEW_HELP_MSG));
    socket_stream_write(cl->socket_stream, NETWORK_NEWLINE, strlen(NETWORK_NEWLINE));
    snprintf(errStr, errStrLen, "#0" NETWORK_NEWLINE);
  }
  else {
    Query q = 0;
    int qid = processQueryInt(query, &q, cl);

    if (qid < 0) {
      if ( q && *(q->error_message) ) {
        snprintf(errStr, errStrLen, "#%i:%s" NETWORK_NEWLINE, qid, q->error_message);
      }
      else {
        snprintf(errStr, errStrLen, "#%i" NETWORK_NEWLINE, qid);
      }
    } else {
      if ( q && *(q->error_message) ) {
        snprintf(errStr, errStrLen, "#0:%s" NETWORK_NEWLINE, q->error_message);
      }
      else {
        snprintf(errStr, errStrLen, "#0" NETWORK_NEWLINE);
      }
    }

    if ( q ) freeQuery(q);
  }

  socket_stream_write(cl->socket_stream, errStr, strlen(errStr));
  socket_stream_flush(cl->socket_stream);

  return 0;
}

static
void
freeRDFValue(RDF_Value v) {
  switch(v.type) {
  case RDF_RESOURCE_TYPE:
  case RDF_STRING_TYPE:
  case RDF_DATE_TYPE: /* This isn't actually used, but if it was used it would probably have to be freed */
  case RDF_BLOB_TYPE: /* This isn't actually used, but if it was used it would probably have to be freed */
  case RDF_VAR_TYPE:
    if ( v.value ) freeMem(v.value);
  }
}

static
void
freeQuery (Query nq) {
  Statement* l = nq->statements;
  Variable v = nq->variables;
  while (v) {
    Variable next = v->next;
    if ( v->name ) freeMem((void*)v->name);
    freeMem((void*)v);
    v = next;
  }
  while (l) {
    Statement* next = l->next;
    freeRDFValue(l->source);
    freeRDFValue(l->arc);
    freeRDFValue(l->target);
    l = next;
  }
  if ( nq->database ) freeMem((void*)nq->database);
  if ( nq->url ) freeMem((void*)nq->url);
  if ( nq->queryString ) freeMem((void*)nq->queryString);
  destroy_constraint_list(nq->constraint_list);
  destroy_for_list(nq->for_list);
  freeMem((void*)nq);
}

void delete_statement(Statement* statement) {
  if ( statement ) {
    freeRDFValue(statement->source);
    freeRDFValue(statement->arc);
    freeRDFValue(statement->target);
  }
}
