/* Copyright (c) 2004, Jasper van de Gronde
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms,
 * with or without modification, are permitted provided
 * that the following conditions are met:
 *
 *  1. Redistributions of source code must retain the above
 *     copyright notice, this list of conditions and the
 *     following disclaimer.
 *  2. Redistributions in binary form must reproduce the
 *     above copyright notice, this list of conditions
 *     and the following disclaimer in the documentation
 *     and/or other materials provided with the distribution.
 *  3. Neither the name of the author nor the names of
 *     its contributors may be used to endorse or promote products
 *     derived from this software without specific prior written
 *     permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR "AS IS" AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY
 * AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT,
 * INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN
 * IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#ifndef RDFDB_QUERY_SELECTPARSER_H
#define RDFDB_QUERY_SELECTPARSER_H

#include "../librdf/rdf.h"
#include "../server/connection.h"
#include <regexp/regexp.h>
#include "../utils/utils.h"

/* This header file includes all the definitions
 * needed for parsing (a modified version of) SquishQL.
 */

struct state_t {
  const char* querystart;
  const char* queryend;
  const char* curstart;
  const char* curend;
  const char* errstart;
  const char* errend;
};

/* These are all stored without quotes or prefixes (in the case of var's). */
typedef char* textliteral_t;
typedef char* uriliteral_t;
typedef int numliteral_t;
typedef char* var_t;
typedef char* db_t;

char* parse_text_literal(struct state_t* state);
char* parse_uri(struct state_t* state);
BOOL parse_numliteral(int* val, struct state_t* state);

struct var_or_literal_t {
  enum {
    VAR_T = RDF_VAR_TYPE,
    NUMLITERAL_T = RDF_INT_TYPE,
    URI_T = RDF_RESOURCE_TYPE,
    TEXTLITERAL_T = RDF_STRING_TYPE,
  } type;
  numliteral_t num; /* only valid when this is a numeric literal */
  char* text; /* represents the literal/var in all other cases */
};
void destroy_var_or_literal(struct var_or_literal_t*);
BOOL parse_var_or_literal(struct var_or_literal_t* vol, struct state_t* state);

/* Var ( '>' | '<' | '==' | '=' | '!=' | <> | '<=' | '>=' ) VarOrLiteral
   Var ( 'like' | 'eq' | 'ne' | '~' ) VarOrLiteral */
struct expression_t {
  var_t var;
  enum {
    STRCOMP_BEGIN = 1,
    STRCOMP_NE,
    STRCOMP_EQ,
    STRCOMP_MATCH,
    STRCOMP_NMATCH,
    STRCOMP_END,

    NUMCOMP_BEGIN = 10,
    NUMCOMP_GT, /* '>' */
    NUMCOMP_LT, /* '<' */
    NUMCOMP_EQ, /* both '==' and '=' */
    NUMCOMP_NE, /* both '!=' and '<>' */
    NUMCOMP_LTEQ, /* '<=' */
    NUMCOMP_GTEQ, /* '>=' */
    NUMCOMP_END,
  } operation;
  struct var_or_literal_t var_or_literal;
  regexp* rx; /* only valid if operation == STRCOMP_MATCH or STRCOMP_NMATCH */
};
void destroy_expression(struct expression_t*);

struct triple_pattern_t {
  struct var_or_literal_t subject;
  struct var_or_literal_t predicate;
  struct var_or_literal_t object;
  int optional;
};
void destroy_triple_pattern(struct triple_pattern_t*);
BOOL parse_triple_pattern(struct triple_pattern_t* tp, struct state_t* state);

/*struct for_list_t {
  struct for_list_t* next;
  char* id;
  char* ns;
};*/
#define for_list_t _NamespaceStruct
void destroy_for_list(struct for_list_t*);
struct for_list_t* parse_for_list(struct state_t* state);

struct constraint_list_t {
  struct constraint_list_t* next;
  struct expression_t expr;
};
void destroy_constraint_list(struct constraint_list_t*);
struct constraint_list_t* parse_constraint_list(struct state_t* state);

struct triple_pattern_list_t {
  struct triple_pattern_list_t* next;
  struct triple_pattern_t tp;
};
void destroy_triple_pattern_list(struct triple_pattern_list_t*);

struct var_list_t {
  struct var_list_t* next;
  var_t var;
};
void destroy_var_list(struct var_list_t*);

/* Represents Query
   select VarList from DB where TriplePatternList (and ConstraintList)? (using ForList)? */
struct select_query_t {
  struct var_list_t* vars;
  db_t db;
  struct triple_pattern_list_t* triple_pattern_list;
  struct constraint_list_t* constraint_list;
  struct for_list_t* for_list;
};
void destroy_select_query(struct select_query_t*);

struct select_query_t* parseSelectQuery(const char* query);

#endif/*RDFDB_QUERY_SELECTPARSER_H*/
