/* Copyright (c) 2004, Jasper van de Gronde
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms,
 * with or without modification, are permitted provided
 * that the following conditions are met:
 *
 *  1. Redistributions of source code must retain the above
 *     copyright notice, this list of conditions and the
 *     following disclaimer.
 *  2. Redistributions in binary form must reproduce the
 *     above copyright notice, this list of conditions
 *     and the following disclaimer in the documentation
 *     and/or other materials provided with the distribution.
 *  3. Neither the name of the author nor the names of
 *     its contributors may be used to endorse or promote products
 *     derived from this software without specific prior written
 *     permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR "AS IS" AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY
 * AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT,
 * INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN
 * IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include "connection.h"
#include "../utils/utils.h"
#include <malloc.h>
#include <pthread.h>
#include <sched.h>
#ifdef WIN32
#include    <winsock2.h>
#endif

static int allow_new_connections = 0;
static unsigned int connection_index = 0;

static pthread_mutex_t connections_mutex = PTHREAD_MUTEX_INITIALIZER;
static ClientConnection connections = 0;

static pthread_attr_t attr;

int ProcessQuery(const char* query,  ClientConnection cl);

static void delete_client_connection(ClientConnection conn);
static void* process_connection(void* param);
/*! Return < 0 on socket error, 0 when the connection has been gracefully closed, 1 if everything is okay or 2 if there wasn't enough space to store the entire query. */
static int read_query(ClientConnection conn, char* query, size_t capacity);

int init_connection() {
  int failures = 0;
  allow_new_connections = 1;
  if ( 0 != pthread_attr_init(&attr) ) ++failures;
  if ( 0 != pthread_attr_setdetachstate(&attr, PTHREAD_CREATE_DETACHED) ) ++failures;
  return failures;
}

int deinit_connection() {
  int failures = 0;
  allow_new_connections = 0;
  pthread_mutex_lock(&connections_mutex);
  {
    ClientConnection conn = connections;
    while(conn) {
      closesocket(conn->raw_socket); /* shutdown doesn't work (at least on Win32) because you have to invoke a read or send (on either side probably) AFTER you have issued the shutdown to make it work. */
      conn->raw_socket = 0;
      conn = conn->next;
    }
  }
  pthread_mutex_unlock(&connections_mutex);
  while(connections) {
    sched_yield();
  }
  pthread_attr_destroy(&attr);
  return failures;
}

ClientConnection new_client_connection(SOCKET socket, unsigned int ipv4address) {
  ClientConnection conn;
  if ( !allow_new_connections || !socket || !ipv4address ) return 0;
  conn = malloc(sizeof(ClientConnectionStruct));
  if ( conn ) {
    pthread_t thread_handle;
    memset(conn, 0, sizeof(ClientConnectionStruct));
    conn->raw_socket = socket;
    conn->security_context = new_security_context(ipv4address);
    if ( !conn->security_context || !action_allowed(conn->security_context, action_connect, 0) ) {
      delete_security_context(conn->security_context);
      closesocket(conn->raw_socket);
      free(conn);
      return 0;
    }
    conn->socket_stream = new_socket_stream(socket, 1024, 1024);
    if ( !conn->socket_stream ) {
      delete_security_context(conn->security_context);
      closesocket(conn->raw_socket);
      free(conn);
      return 0;
    }
    
    pthread_mutex_lock(&connections_mutex);
    conn->next = connections;
    connections = conn;
    pthread_mutex_unlock(&connections_mutex);

    conn->index = connection_index;

    if ( 0 != pthread_create(&thread_handle, &attr, process_connection, conn) ) {
      delete_client_connection(conn);
      return 0;
    }

    ++connection_index;
  }
  return conn;
}

static void delete_client_connection(ClientConnection conn) {
  if ( conn ) {
    pthread_mutex_lock(&connections_mutex);
    {
      ClientConnection prev = 0;
      ClientConnection cur = connections;
      while(cur) {
        if ( cur == conn ) {
          if ( prev ) {
            prev->next = cur->next;
          }
          else {
            connections = cur->next;
          }
          break;
        }
        prev = cur;
        cur = cur->next;
      }
    }
    pthread_mutex_unlock(&connections_mutex);

    delete_socket_stream(conn->socket_stream);
    delete_security_context(conn->security_context);
    if ( conn->raw_socket ) closesocket(conn->raw_socket);
    {
      Namespace ns = conn->namespaces;
      while(ns) {
        Namespace next = ns->next;
        freeNamespace(ns);
        ns = next;
      }
    }

    free(conn);
  }
}

static void* process_connection(void* param) {
  ClientConnection conn = param;
  char query[4096];
  int err;

  PrintMsg(INFO, "Connection %u opened.\n", conn->index);

  for(;;) {
    err = read_query(conn, query, sizeof(query));
    if ( err == 1 ) {
      ProcessQuery(query, conn);
    }
    else if ( err == 2 ) {
      const char* message = "%-4::Query too long" NETWORK_NEWLINE;
      socket_stream_write(conn->socket_stream, message, strlen(message));
      socket_stream_flush(conn->socket_stream);
    }
    else {
      break;
    }
  }

  if ( err == 0 ) {
    PrintMsg(INFO, "Connection %u closed gracefully.\n", conn->index);
  }
  else if ( err < 0 ) {
    PrintMsg(INFO, "Connection %u aborted.\n", conn->index);
  }
  else {
    PrintMsg(INFO, "Connection %u aborted (unrecognized server-side error).\n", conn->index);
  }

  delete_client_connection(conn);

  return 0;
}

static int read_query(ClientConnection conn, char* query, size_t capacity) {
  while(capacity) {
    size_t read = socket_stream_readline(conn->socket_stream, query, capacity);
    if ( read <= 0 ) {
      return read;
    }
    else if ( read == capacity ) {
      char buffer[1024];
      do {
        read = socket_stream_readline(conn->socket_stream, buffer, sizeof(buffer));
      } while(read == capacity);
      return 2;
    }

    read -= inplace_trim_line(query);

    if ( endsWith(query, "</>") ) {
      read -= 3;
      while(read && isspace(*(query+read-1))) --read;
      *(query+read) = 0;
      break;
    }

    query += read;
    capacity -= read;
  }
  return 1;
}
