/* -*- Mode: C; tab-width: 4; indent-tabs-mode: nil; c-basic-offset: 2 -*-
 *
 * The contents of this file are subject to the Mozilla Public 
 * License version 1.1 (the "License"); you may not use this file
 * except in compliance with the License. You may obtain a copy of
 * the License at http://www.mozilla.org/MPL/
 *
 * Software distributed under the License is distributed on an "AS
 * IS" basis, WITHOUT WARRANTY OF ANY KIND, either express or 
 * implied. See the License for the specific language governing
 * rights and limitations under the License.
 *
 * The Initial Developer of the Original Code is R.V.Guha. Portions
 * created by R.V.Guha are copyright (c) 2000 R.V.Guha. All Rights
 * Reserved.
 */

#include "hash.h"
#include "utils.h"

int 
hashKey (HashTable ht, const char* key) {
  size_t len = strlen(key);
  int    sum = 0;
  size_t    n = 0;
  int ans;
  for (n = 0; n < len; n++) sum = sum + (int)key[n];
  ans = sum & (ht->size - 1); /* JvdG Changed added the -1, only works correctly if ht-size is an integer power of two. */
  if (ans == ht->size) ans = ans-1;
  return ans;
}

HashTable 
NewHashTable(int size) {
  HashTable ht;
  { /* JvdG Added this bit to make sure the size is a power of two. */
    int s = 1;
    while(s<size) s<<=1; /* This will fail if size > 2^31, which is unlikely */
    if ( s != size ) {
      return 0;
    }
  }
  ht = (HashTable)getMem(sizeof(HashTableStruct));
  if ( ht ) {
    ht->size = size;
    ht->buckets = (HashEntry*)getMem(sizeof(HashEntry) * size);
  }
  return ht;
}

/* JvdG Renamed this to DeleteHashTable (it used to be DestroyHashTable) and made it use the delete convention of not doing anything if ht == 0. */
void
DeleteHashTable(HashTable ht) {
  if ( ht ) {
    int i;
    for(i = 0; i < ht->size; ++i) {
      HashEntry next;
      HashEntry he = ht->buckets[i];
      if ( he ) he = he->next;
      while(he) {
        free(he->key); /* values aren't the HashTable's responsibility */
        next = he->next;
        freeMem(he);
        he = next;
      }
    }
    freeMem(ht->buckets);
    freeMem(ht);
  }
}

/* JvdG Made this use the order of the entries. */
void*
HashLookup(HashTable ht, const char* key) {
  int offset = hashKey(ht, key);
  int cmp = 1;
  HashEntry he = ht->buckets[offset];
  while (he && (cmp=strcmp(he->key, key)) < 0) {
    he = he->next;
  }
  if ( cmp == 0 ) return he->value;
  return NULL;
}

/* JvdG Made this insert new entries in order. */
void* 
HashAdd (HashTable ht, const char* key, void* value) {
  int offset = hashKey(ht, key);
  int cmp = 1;
  HashEntry he = ht->buckets[offset];
  HashEntry prev = 0;
  while (he && (cmp=strcmp(he->key, key)) < 0) {
    prev = he;
    he = he->next;
  }
  if ( cmp == 0) {
    if (value == he->value) {
	    return value;
    } else {
      void* old_value = he->value;
	    he->value = value;
	    return old_value;
    }
  }
  else {
    HashEntry newhe = (HashEntry) getMem(sizeof(HashEntryStruct));
    newhe->next  = he;
    newhe->value = value;
    newhe->key   = strdup(key);
    if (prev) {
      prev->next = newhe;
    } else {
      ht->buckets[offset] = newhe;
    }
    return 0;
  }
}
 
