/* Copyright (c) 2004, Jasper van de Gronde
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms,
 * with or without modification, are permitted provided
 * that the following conditions are met:
 *
 *  1. Redistributions of source code must retain the above
 *     copyright notice, this list of conditions and the
 *     following disclaimer.
 *  2. Redistributions in binary form must reproduce the
 *     above copyright notice, this list of conditions
 *     and the following disclaimer in the documentation
 *     and/or other materials provided with the distribution.
 *  3. Neither the name of the author nor the names of
 *     its contributors may be used to endorse or promote products
 *     derived from this software without specific prior written
 *     permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR "AS IS" AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY
 * AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT,
 * INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN
 * IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include "socketstream.h"
#include <malloc.h>
#ifdef WIN32
#include    <winsock2.h>
#endif

struct _socket_stream_t {
  SOCKET socket;

  char* read_buffer;
  int read_buffer_size;
  int read_buffer_beginpos;
  int read_buffer_endpos;

  char* write_buffer;
  int write_buffer_size;
  int write_buffer_pos;
};

#ifdef WIN32
socket_stream_t* new_socket_stream(SOCKET socket, int read_buf_size, int write_buf_size)
#else
socket_stream_t* new_socket_stream(int socket, int read_buf_size, int write_buf_size)
#endif
{
  socket_stream_t* socket_stream = malloc(sizeof(socket_stream_t));
  if ( socket_stream ) {
    socket_stream->socket = socket;
    socket_stream->read_buffer_beginpos = 0;
    socket_stream->read_buffer_endpos = 0;
    socket_stream->read_buffer_size = read_buf_size;
    socket_stream->write_buffer_pos = 0;
    socket_stream->write_buffer_size = write_buf_size;
    socket_stream->read_buffer = malloc(read_buf_size);
    if ( !socket_stream->read_buffer ) {
      free(socket_stream);
      return 0;
    }
    socket_stream->write_buffer = malloc(write_buf_size);
    if ( !socket_stream->write_buffer ) {
      free(socket_stream->read_buffer);
      free(socket_stream);
      return 0;
    }
  }
  return socket_stream;
}

void delete_socket_stream(socket_stream_t* socket_stream) {
  if ( socket_stream ) {
    free(socket_stream->read_buffer);
    free(socket_stream->write_buffer);
    free(socket_stream);
  }
}

/* Should return the number of bytes sent, or a value below 0 to indicate an error. */
#ifdef WIN32
static int socket_stream_write_raw(socket_stream_t* socket_stream, const void* data, int data_size) {
  return send(socket_stream->socket, data, data_size, 0);
}
#else
static int socket_stream_write_raw(socket_stream_t* socket_stream, const void* data, int data_size) {
  return write(socket_stream->socket, data, data_size, 0);
}
#endif

/* Should return the number of bytes sent, 0 to indicate a gracefully closed socket or a value below 0 to indicate an error. */
#ifdef WIN32
static int socket_stream_read_raw(socket_stream_t* socket_stream, void* data, int data_capacity) {
  return recv(socket_stream->socket, data, data_capacity, 0);
}
#else
static int socket_stream_read_raw(socket_stream_t* socket_stream, void* data, int data_capacity) {
  return read(socket_stream->socket, data, data_capacity);
}
#endif

int socket_stream_write(socket_stream_t* socket_stream, const void* data_p, int data_size) {
  
  const char* data = data_p;
  char* const buffer = socket_stream->write_buffer;
  const int max_chunk_size = socket_stream->write_buffer_size;
  const int curpos = socket_stream->write_buffer_pos;
  int to_write = data_size;
  int chunk_size = to_write > (max_chunk_size-curpos) ? (max_chunk_size-curpos) : to_write;
  memcpy(buffer+curpos, data, chunk_size);
  socket_stream->write_buffer_pos += chunk_size;
  to_write -= chunk_size;
  while(to_write) {
    if ( socket_stream_flush(socket_stream) < 0 ) {
      return -1;
    }
    data += chunk_size;
    chunk_size = to_write > max_chunk_size ? max_chunk_size : to_write;
    memcpy(buffer, data, chunk_size);
    socket_stream->write_buffer_pos += chunk_size;
    to_write -= chunk_size;
  }
  return data_size-to_write;
  /*
  const char* data = data_p;
  int to_write = data_size;
  while(to_write) {
    const int written = socket_stream_write_raw(socket_stream, data, to_write);
    if ( written < 0 ) {
      return -1;
    }
    data += written;
    to_write -= written;
  }
  return data_size;
  */
}
/*
int socket_stream_read(socket_stream_t* socket_stream, void* data_p, int data_capacity) {
  char* data = data_p;
  char* const buffer = socket_stream->read_buffer;
  const int max_chunk_size = socket_stream->read_buffer_size;
  const int beginpos = socket_stream->read_buffer_beginpos;
  const int endpos = socket_stream->read_buffer_endpos;
  int to_read = data_capacity;
  int chunk_size = to_read > (endpos-beginpos) ? (endpos-beginpos) : to_read;
  memcpy(data, buffer+beginpos, chunk_size);
  to_read -= chunk_size;
  while(to_read) {
    socket_stream->read_buffer_beginpos = 0;
    socket_stream->read_buffer_endpos = socket_stream_read_raw(socket_stream, socket_stream->read_buffer, socket_stream->read_buffer_size);
    if ( socket_stream->read_buffer_endpos <= 0 ) {
      socket_stream->read_buffer_endpos = 0;
      return -1;
    }
    data += chunk_size;
    chunk_size = to_read > socket_stream->read_buffer_endpos ? socket_stream->read_buffer_endpos : to_read;
    memcpy(data, buffer, chunk_size);
    to_read -= chunk_size;
  }
  socket_stream->read_buffer_beginpos += chunk_size;
  return data_capacity-to_read;
}
*/
int socket_stream_flush(socket_stream_t* socket_stream) {
  char* buffer = socket_stream->write_buffer;
  int to_write = socket_stream->write_buffer_pos;
  int written_total = 0;
  while(to_write) {
    const int written = socket_stream_write_raw(socket_stream, buffer, to_write);
    if ( written < 0 ) {
      memmove(buffer, buffer+written_total, to_write);
      socket_stream->write_buffer_pos = to_write;
      return -1;
    }
    written_total += written;
    buffer += written;
    to_write -= written;
  }
  socket_stream->write_buffer_pos = 0;
  return written_total;
}

int socket_stream_readline(socket_stream_t* socket_stream, void* data_p, int data_capacity) {
  char* data = data_p;
  char* buffer = socket_stream->read_buffer;
  int beginpos = socket_stream->read_buffer_beginpos;
  int endpos = socket_stream->read_buffer_endpos;
  int to_read = data_capacity - 1;
  int newlinepos;
  while(to_read) {
    int chunk_size;
    newlinepos = beginpos;
    while(newlinepos < endpos && buffer[newlinepos] != '\xA' ) {
      ++newlinepos;
    }
    chunk_size = newlinepos-beginpos;
    if ( chunk_size > to_read ) chunk_size = to_read;
    memcpy(data, buffer+beginpos, chunk_size);
    data += chunk_size;
    to_read -= chunk_size;
    if ( beginpos+chunk_size < newlinepos ) {
      newlinepos = beginpos+chunk_size;
      to_read = -1;
      break;
    }
    else if ( newlinepos != endpos ) {
      if ( !(data != data_p && *(data-1) == '\xD') ) {
        ++newlinepos;
      }
      break;
    }
    else if ( to_read ) {
      beginpos = 0;
      endpos = socket_stream_read_raw(socket_stream, socket_stream->read_buffer, socket_stream->read_buffer_size);
      if ( endpos <= 0 ) {
        socket_stream->read_buffer_endpos = 0;
        return endpos;
      }
    }
  }
  socket_stream->read_buffer_beginpos = newlinepos;
  socket_stream->read_buffer_endpos = endpos;
  *data = 0;
  return (data_capacity-1)-to_read;
}
